import numpy as np
from tqdm import tqdm
from scipy.spatial import cKDTree as KDTree
from scipy.sparse import csgraph

DEFAULT_BINS_GRAPH = np.append(np.arange(0, 15), np.inf)
DEFAULT_P_RATIO = np.array([1.66337261,  1.47190522,  1.60522673,  1.27751097,  0.73914456,
                            -0.0630327, -1.44133147, -2.25994548, -2.77429373, -2.77863396,
                            -4.52941582, -4.68640886, -5.67060111, -6.41176085, -9.43203527])


def skeleton_path_quality(sk, mesh, skind_to_mind_map=None, sk_norm=None, max_graph_dist=10000,
                          p_ratio=None, bins=None, normalize_to_self=True,
                          sliding=True, window=400, interval=50, return_path_info=False, filter_skind_map=True):
    '''
    Compute path quality for all endpoints of a skeletonization by comparing paths from skeleton tips to root to
    the shortest path between the same points along the mesh.
    :param sk: MeshParty Skeleton, derived from the mesh also given as an argument. The skeleton has N_s vertices.
    :param mesh: MeshParty Mesh. The mesh has N_m vertices.
    :param skind_to_mind_map: N_s length array or list. Optional, default is None. Each the value of the ith
                entry is the index of the mesh vertex corresponding to the ith skeleton vertex. If none is given,
                the value defaults to sk.vertex_properties['mesh_index'], created by default in skeletonization.skeletonize_mesh.
    :param sk_norm: Float or N_s length array or list. Optional, default is None. A normalization value for the distance
                between paths to account for different-width neurites. If None is given, the value defaults to 
                sk.radius, the mesh radius value created by default in skeletonization.skeletonize_mesh.
                If the radius is used, the pre-computed default likelihood ratio/binning.
    :param max_graph_dist: Float. Optional, default=10000. Values behind this are treated as infinity in distance computations.
    :param p_ratio: List with M values. Optional, defualt is None. Denotes the log likelihood ratio that a given distance bin came from a matched
                path pair vs an unmatched path pair.  If None is given, it defaults to the value DEFAULT_P_RATIO above, trained on
                a collection of paths from seven pyramidal neurons in the Pinky100 dataset.
    :param bins: List with M+1 values. Optional, default is None. Gives the bin edges for the distances corresponding to the
                p_ratio values. If None is given, it defaults to DEFAULT_BINS_GRAPH.
    :param normalize_to_self: Boolean. Optional, default is True. If False, the sum of log likelihoods for path pairs is given.
                If True, the values are normalized to the best-case scenario of the log-likelihood distane of the skeleton path to itself.
    :param sliding: Boolean. Optional, default is True. If True, takes the worst score along a sliding window along the paths to avoid
                path length diluting a small misalignment between paths. If False, the whole paths are used.
    :param window: Int. Optional, default is 400. The path length (in hops) of the window to use for the sliding quality measure.
    :param interval: Int. Optional, default is 50. The interval between window starts in teh sliding quality measure.
    :param return_path_info: Boolean. Optional, default is False. If True, returns a list of path_quality values and the paths themselves.
                If False, returns only the path quality.
    :param filter_skind_map: Boolean. Optional, default is True. If True, assumes the skind_to_mind_map is in the indices of the original
                mesh, not a filtered version. Otherwise, ignores the node_mask. Should be True if using default values for skind_to_mind_map.
    :returns: path_quality (if return_path_info is False)
              path_quality, sk_paths, ms_paths, sk_inds_list, mesh_inds_list, path_distances (if return_path_info is True)
    '''
    if skind_to_mind_map is None:
        ds, inds = mesh.kdtree.query(sk.vertices, distance_upper_bound=1)
        skind_to_mind_map = mesh.map_indices_to_unmasked(
            np.where(ds == 0, inds, -1))
    if sk_norm is None:
        if sk.radius is not None:
            sk_norm = sk.radius
        else:
            sk_norm = np.ones(sk.n_vertices)
    elif p_ratio is not None:
        print(
            'Warning: If sk_norm is not the radius, the default calibration does not apply!')

    if p_ratio is None:
        p_ratio = DEFAULT_P_RATIO

    if bins is None:
        bins = DEFAULT_BINS_GRAPH

    if type(skind_to_mind_map) is list:
        skind_to_mind_map = np.array(skind_to_mind_map)
    if type(sk_norm) is list:
        sk_norm = np.array(sk_norm)
    if filter_skind_map is True:
        skind_to_mind_map = mesh.filter_unmasked_indices(skind_to_mind_map)
    print("Computing path pairs...")
    sk_paths, ms_paths, sk_inds_list, mesh_inds_list = compute_path_pairs(
        sk, mesh, skind_to_mind_map=skind_to_mind_map)
    print("Computing matched path distances...")
    path_distances = matched_path_distances_mesh(sk_inds_list, mesh_inds_list, skind_to_mind_map, mesh,
                                                 sk_norm=sk_norm, max_dist=max_graph_dist)
    path_quality = []
    for pd in path_distances:
        if sliding:
            path_quality.append(pblast_score_sliding(pd, p_ratio=p_ratio,
                                                     bins=bins, normalize_to_self=normalize_to_self))
        else:
            path_quality.append(pblast_score(pd, p_ratio=p_ratio, bins=bins,
                                             normalize_to_self=normalize_to_self))
    if return_path_info:
        return np.array(path_quality), sk_paths, ms_paths, sk_inds_list, mesh_inds_list, path_distances
    else:
        return np.array(path_quality)


def compute_path_pairs(sk, mesh, skind_to_mind_map=None, return_indices=True):
    '''
    For every end point on a skeleton, computes two shortest paths to root,
    one along the skeleton and one along the mesh.
    :param sk: Skeleton
    :param mesh: MeshParty Mesh
    :param skind_to_mind_map: Optional, Mapping from index in skeleton to index in the mesh.
                              Usually 'mesh_index' in skeleton vertex properties.
    :param return_indices: Optional, boolean. If included, also return indices, not just points.
    :returns: sk_paths ms_paths, sk_path_indices, ms_path_indices
    '''
    end_points = sk.end_points
    mesh_to_skel_map = sk.mesh_to_skel_map[mesh.node_mask]
    root_minds = np.flatnonzero(mesh_to_skel_map == sk.root)

    # Compute a minimal set of indices to use in the dijkstra to avoid extra computation if all soma indices are included.
    mesh_vec = np.zeros((len(mesh.vertices), 1))
    mesh_vec[root_minds] = 1
    meshgraph = mesh.csgraph > 0
    soma_boundary_plus_vec = (
        ((meshgraph * mesh_vec) > 0).astype(int) - mesh_vec).astype(int)
    soma_boundary_inds = np.flatnonzero(np.logical_and(
        (meshgraph * soma_boundary_plus_vec > 0), mesh_vec > 0))

    ds, Ps, _ = csgraph.dijkstra(mesh.csgraph, return_predecessors=True,
                                 indices=soma_boundary_inds, min_only=True)
    sk_paths = []
    ms_paths = []
    if return_indices:
        ms_path_indices = []
        sk_path_indices = []
    for sk_ep in tqdm(sk.end_points):
        rel_minds = np.flatnonzero(mesh_to_skel_map == sk_ep)
        if skind_to_mind_map is None:
            rel_minds = np.flatnonzero(mesh_to_skel_map == sk_ep)
            ms_ep = rel_minds[np.argmin(np.linalg.norm(
                mesh.vertices[rel_minds] - sk.vertices[sk_ep], axis=1))]
        else:
            ms_ep = skind_to_mind_map[sk_ep]
        ms_path_inds = []
        ms_ind = ms_ep
        while ms_ind != -9999:  # Set by csgraph.dijkstra
            ms_path_inds.append(ms_ind)
            ms_ind = Ps[ms_ind]

        # Don't include root, since the mesh path won't have it
        sk_path = sk.vertices[sk.path_to_root(sk_ep)[:-1]]
        # Don't include root, since the mesh path won't have it
        ms_path = mesh.vertices[ms_path_inds]
        if len(sk_path) > 0:
            sk_paths.append(sk_path)
            ms_paths.append(ms_path)
            if return_indices:
                sk_path_indices.append(np.array(sk.path_to_root(sk_ep)[:-1]))
                ms_path_indices.append(np.array(ms_path_inds))
    if return_indices:
        return sk_paths, ms_paths, sk_path_indices, ms_path_indices
    else:
        return sk_paths, ms_paths


def matched_path_distances_normalized(sk_paths, mesh_paths, sk_inds_list, sk_rs, r_min=50, from_path='skel'):
    '''
    Compute euclidean distances between all paired paths.
    '''
    path_distances = []
    for sk_path, mesh_path, sk_inds in zip(sk_paths, mesh_paths, sk_inds_list):
        if from_path == 'mesh':
            ds, c_skinds = KDTree(sk_path).query(mesh_path)
            ds_norm = ds / np.clip(sk_rs[c_skinds], r_min, None)
        elif from_path == 'skel':
            ds, c_minds = KDTree(mesh_path).query(sk_path)
            ds_norm = ds / np.clip(sk_rs[sk_inds], r_min, None)
        path_distances.append(ds_norm)
    return path_distances


def matched_path_distances_mesh(sk_inds_list, mesh_inds_list, skind_to_mind_map, mesh, sk_norm=None, max_dist=10000):
    '''
    Compute on-mesh distance from skeleton to mesh path between all paired paths.
    Optionally, normalize distance by a value for every skeleton point, typically the radius.
    '''
    path_distances = []
    for skind_path, mind_path in tqdm(zip(sk_inds_list, mesh_inds_list)):
        ds = closest_graph_distance_paths(skind_to_mind_map[skind_path], mind_path, mesh,
                                          norm=sk_norm[skind_path], max_dist=max_dist)
        path_distances.append(ds)
    return path_distances


def closest_graph_distance_paths(path_A, path_B, mesh, norm=None, norm_min=50, max_dist=10000):
    ds = csgraph.dijkstra(mesh.csgraph, indices=path_B,
                          min_only=True, limit=max_dist)
    if norm is None:
        return ds[path_A]
    else:
        return ds[path_A] / np.clip(norm, norm_min, None)


def pblast_score(data, p_ratio=DEFAULT_P_RATIO, bins=DEFAULT_BINS_GRAPH, normalize_to_self=True):
    '''
    For a set of path distances, compute a metric that is basically a naive Bayes likelihood ratio test
    between being from a 'good' skeletonization (positive numbers) and a path that uses a bad merge
    (negative numbers).
    '''
    d_ns, _ = np.histogram(data, bins=bins)
    if normalize_to_self is True:
        norm = np.sum(len(data) * p_ratio[0])
    else:
        norm = 1
    return np.sum(np.dot(p_ratio, d_ns)) / norm


def pblast_score_sliding(data, window=400, interval=50, p_ratio=DEFAULT_P_RATIO, bins=DEFAULT_BINS_GRAPH, normalize_to_self=True):
    if len(data) < window:
        return pblast_score(data, p_ratio=p_ratio, bins=bins, normalize_to_self=normalize_to_self)
    else:
        worst_pscore = np.inf
        for ii in np.arange(0, len(data)-window+interval, interval):
            new_pscore = pblast_score(data[ii:ii+window], p_ratio=p_ratio,
                                      bins=bins, normalize_to_self=normalize_to_self)
            if new_pscore < worst_pscore:
                worst_pscore = new_pscore
        return worst_pscore



from scipy import sparse, spatial, optimize, signal
import numpy as np
import time
from meshparty import utils

try:
    from pykdtree.kdtree import KDTree
except:
    KDTree = spatial.cKDTree
from tqdm import tqdm
from meshparty.skeleton import Skeleton
from .ray_tracing import ray_trace_distance, shape_diameter_function
import fastremap
import logging
from . import skeleton_utils


def skeletonize_mesh(
    mesh,
    soma_pt=None,
    soma_radius=7500,
    collapse_soma=True,
    collapse_function="sphere",
    invalidation_d=12000,
    smooth_vertices=False,
    compute_radius=True,
    shape_function="single",
    compute_original_index=True,
    verbose=True,
    smooth_iterations=12,
    smooth_neighborhood=2,
    smooth_r=0.1,
    cc_vertex_thresh=100,
    root_index=None,
    remove_zero_length_edges=True,
    collapse_params={},
    meta={},
):
    """
    Build skeleton object from mesh skeletonization

    Parameters
    ----------
    mesh: meshparty.trimesh_io.Mesh
        the mesh to skeletonize, defaults assume vertices in nm
    soma_pt: np.array
        a length 3 array specifying to soma location to make the root
        default=None, in which case a heuristic root will be chosen
        in units of mesh vertices.
    soma_radius: float
        distance in mesh vertex units over which to consider mesh
        vertices close to soma_pt to belong to soma
        these vertices will automatically be invalidated and no
        skeleton branches will attempt to reach them.
        This distance will also be used to collapse all skeleton
        points within this distance to the soma_pt root if collpase_soma
        is true. (default=7500 (nm))
    collapse_soma: bool
        whether to collapse the skeleton around the soma point (default True)
    collapse_function: 'sphere' or 'branch'
        Determines which soma collapse function to use. Sphere uses the soma_radius
        and collapses all vertices within that radius to the soma. Branch is an experimental
        approach that tries to compute the right boundary for each branch into soma.
    invalidation_d: float
        the distance along the mesh to invalidate when applying TEASAR
        like algorithm.  Controls how detailed a structure the skeleton
        algorithm reaches. default (12000 (nm))
    smooth_vertices: bool
        whether to smooth the vertices of the skeleton
    compute_radius: bool
        whether to calculate the radius of the skeleton at each point on the skeleton
        (default True)
    shape_function: 'single' or 'cone'
        Selects how to compute the radius, either with a single ray or a cone of rays. Default is 'single'.
    compute_original_index: bool
        whether to calculate how each of the mesh nodes maps onto the skeleton
        (default True)
    smooth_iterations: int, optional
        Number of iterations to smooth (default is 12)
    smooth_neighborhood: int, optional
        Size of neighborhood to look at for smoothing
    smooth_r: float, optional
        Weight of update step in smoothing algorithm, default is 0.2
    root_index: int or None, optional
        A precise mesh vertex to use as the skeleton root. If provided, the vertex location overrides soma_pt. By default, None.
    remove_zero_length_edges: bool
        If True, removes vertices involved in zero length edges, which can disrupt graph computations. Default True.
    collapse_params: dict
        Extra keyword arguments for the collapse function. See soma_via_sphere and soma_via_branch_starts for specifics.
    cc_vertex_thresh : int, optional
        Smallest number of vertices in a connected component to skeletonize.
    verbose: bool
        whether to print verbose logging
    meta: dict
        Skeletonization metadata to add to the skeleton. See skeleton.SkeletonMetadata for keys.

    Returns
    -------
    :obj:`meshparty.skeleton.Skeleton`
           a Skeleton object for this mesh
    """
    (
        skel_verts,
        skel_edges,
        orig_skel_index,
        skel_map,
    ) = calculate_skeleton_paths_on_mesh(
        mesh,
        invalidation_d=invalidation_d,
        cc_vertex_thresh=cc_vertex_thresh,
        root_index=root_index,
        return_map=True,
    )

    if smooth_vertices is True:
        smooth_verts = smooth_graph(
            skel_verts,
            skel_edges,
            neighborhood=smooth_neighborhood,
            iterations=smooth_iterations,
            r=smooth_r,
        )
        skel_verts = smooth_verts

    if root_index is not None and soma_pt is None:
        soma_pt = mesh.vertices[root_index]

    if soma_pt is not None:
        soma_pt = np.array(soma_pt).reshape(1, 3)

    rs = None

    if collapse_soma is True and soma_pt is not None:
        temp_sk = Skeleton(
            skel_verts,
            skel_edges,
            mesh_index=mesh.map_indices_to_unmasked(orig_skel_index),
            mesh_to_skel_map=skel_map,
        )
        _, close_ind = temp_sk.kdtree.query(soma_pt)
        temp_sk.reroot(close_ind[0])

        if collapse_function == "sphere":
            soma_verts, soma_r = soma_via_sphere(
                soma_pt, temp_sk.vertices, temp_sk.edges, soma_radius
            )
        elif collapse_function == "branch":

            if shape_function == "single":
                rs = ray_trace_distance(
                    mesh.filter_unmasked_indices_padded(temp_sk.mesh_index), mesh
                )
            elif shape_function == "cone":
                rs = shape_diameter_function(
                    mesh.filter_unmasked_indices_padded(temp_sk.mesh_index),
                    mesh,
                    num_points=30,
                    cone_angle=np.pi / 3,
                )

            soma_verts, soma_r = soma_via_branch_starts(
                temp_sk,
                mesh,
                soma_pt,
                rs,
                search_radius=collapse_params.get("search_radius", 25000),
                fallback_radius=collapse_params.get("fallback_radius", soma_radius),
                cutoff_threshold=collapse_params.get("cutoff_threshold", 0.4),
                min_cutoff=collapse_params.get("min_cutoff", 0.1),
                dynamic_range=collapse_params.get("dynamic_range", 1),
                dynamic_threshold=collapse_params.get("dynamic_threshold", False),
            )
        if root_index is not None:
            collapse_index = np.flatnonzero(orig_skel_index == root_index)[0]
        else:
            collapse_index = None
        new_v, new_e, new_skel_map, vert_filter, root_ind = collapse_soma_skeleton(
            soma_verts,
            soma_pt,
            temp_sk.vertices,
            temp_sk.edges,
            mesh_to_skeleton_map=temp_sk.mesh_to_skel_map,
            collapse_index=collapse_index,
            return_filter=True,
            return_soma_ind=True,
        )
    else:
        new_v, new_e, new_skel_map = skel_verts, skel_edges, skel_map
        vert_filter = np.arange(len(orig_skel_index))
        if root_index is not None:
            root_ind = np.flatnonzero(orig_skel_index == root_index)[0]
        elif soma_pt is None:
            sk_graph = utils.create_csgraph(new_v, new_e)
            root_ind = utils.find_far_points_graph(sk_graph)[0]
        else:
            # Still try to root close to the soma
            _, qry_inds = spatial.cKDTree(new_v, balanced_tree=False).query(
                soma_pt[np.newaxis, :]
            )
            root_ind = qry_inds[0]

    skel_map_full_mesh = np.full(mesh.node_mask.shape, -1, dtype=np.int64)
    skel_map_full_mesh[mesh.node_mask] = new_skel_map
    ind_to_fix = mesh.map_boolean_to_unmasked(np.isnan(new_skel_map))
    skel_map_full_mesh[ind_to_fix] = -1

    props = {}

    if compute_original_index is True:
        if collapse_soma is True and soma_pt is not None:
            mesh_index = temp_sk.mesh_index[vert_filter]
            if root_index is None:
                mesh_index = np.append(mesh_index, -1)
        else:
            mesh_index = orig_skel_index[vert_filter]
        props["mesh_index"] = mesh_index

    if compute_radius is True:
        if rs is None:
            if shape_function == "single":
                rs = ray_trace_distance(orig_skel_index[vert_filter], mesh)
            elif shape_function == "cone":
                rs = shape_diameter_function(orig_skel_index[vert_filter], mesh)
        else:
            rs = rs[vert_filter]
        if collapse_soma is True and soma_pt is not None:
            if root_index is None:
                rs = np.append(rs, soma_r)
            else:
                rs[root_ind] = soma_r
        props["rs"] = rs

    sk_params = {
        "soma_pt_x": soma_pt[0, 0] if soma_pt is not None else None,
        "soma_pt_y": soma_pt[0, 1] if soma_pt is not None else None,
        "soma_pt_z": soma_pt[0, 2] if soma_pt is not None else None,
        "soma_radius": soma_radius,
        "collapse_soma": collapse_soma,
        "collapse_function": collapse_function,
        "invalidation_d": invalidation_d,
        "smooth_vertices": smooth_vertices,
        "compute_radius": compute_radius,
        "shape_function": shape_function,
        "smooth_iterations": smooth_iterations,
        "smooth_neighborhood": smooth_neighborhood,
        "smooth_r": smooth_r,
        "cc_vertex_thresh": cc_vertex_thresh,
        "remove_zero_length_edges": remove_zero_length_edges,
        "collapse_params": collapse_params,
        "timestamp": time.time(),
    }
    sk_params.update(meta)

    sk = Skeleton(
        new_v,
        new_e,
        mesh_to_skel_map=skel_map_full_mesh,
        mesh_index=props.get("mesh_index", None),
        radius=props.get("rs", None),
        root=root_ind,
        remove_zero_length_edges=remove_zero_length_edges,
        meta=sk_params,
    )

    if compute_radius is True:
        _remove_nan_radius(sk)

    return sk


def _remove_nan_radius(sk, set_unfixed_to_lowest=True):

    last_numnans = np.inf
    nanlocs = np.flatnonzero(np.isnan(sk.radius))
    numnans = len(nanlocs)
    while numnans > 0 and last_numnans > numnans:
        for nanloc in nanlocs:
            sparse_row = sk.csgraph_binary_undirected[nanloc].toarray().ravel()
            prod = sparse_row * sk.radius
            with np.errstate(divide="ignore", invalid="ignore"):
                new_rad = np.nansum(prod) / np.nansum(prod > 0)
            sk._rooted.radius[nanloc] = new_rad
        last_numnans = numnans
        nanlocs = np.flatnonzero(np.isnan(sk.radius))
        numnans = len(nanlocs)

    if numnans > 0 and set_unfixed_to_lowest:
        sk._rooted.radius[nanlocs] = np.nanmin(sk.radius)


def calculate_skeleton_paths_on_mesh(
    mesh,
    soma_pt=None,
    soma_thresh=7500,
    invalidation_d=10000,
    smooth_neighborhood=2,
    smooth_iterations=12,
    cc_vertex_thresh=100,
    large_skel_path_threshold=5000,
    return_map=False,
    root_index=None,
):
    """function to turn a trimesh object of a neuron into a skeleton, without running soma collapse,
    or recasting result into a Skeleton.  Used by :func:`meshparty.skeletonize.skeletonize_mesh` and
    makes use of :func:`meshparty.skeletonize.skeletonize_components`

    Parameters
    ----------
    mesh: meshparty.trimesh_io.Mesh
        the mesh to skeletonize, defaults assume vertices in nm
    soma_pt: np.array
        a length 3 array specifying to soma location to make the root
        default=None, in which case a heuristic root will be chosen
        in units of mesh vertices
    soma_thresh: float
        distance in mesh vertex units over which to consider mesh
        vertices close to soma_pt to belong to soma
        these vertices will automatically be invalidated and no
        skeleton branches will attempt to reach them.
        This distance will also be used to collapse all skeleton
        points within this distance to the soma_pt root if collpase_soma
        is true. (default=7500 (nm))
    invalidation_d: float
        the distance along the mesh to invalidate when applying TEASAR
        like algorithm.  Controls how detailed a structure the skeleton
        algorithm reaches. default (10000 (nm))
    smooth_neighborhood: int
        the neighborhood in edge hopes over which to smooth skeleton locations.
        This controls the smoothing of the skeleton
        (default 5)
    large_skel_path_threshold: int
        the threshold in terms of skeleton vertices that skeletons will be
        nominated for tip merging.  Smaller skeleton fragments
        will not be merged at their tips (default 5000)
    cc_vertex_thresh: int
        the threshold in terms of vertex numbers that connected components
        of the mesh will be considered for skeletonization. mesh connected
        components with fewer than these number of vertices will be ignored
        by skeletonization algorithm. (default 100)
    return_map: bool
        whether to return a map of how each mesh vertex maps onto each skeleton vertex
        based upon how it was invalidated.
    root_index: int or None
        Mesh vertex to set as initial root node. Overides soma_pt if provided. Default is None.

    Returns
    -------
    skel_verts: np.array
        a Nx3 matrix of skeleton vertex positions
    skel_edges: np.array
        a Kx2 matrix of skeleton edge indices into skel_verts
    smooth_verts: np.array
        a Nx3 matrix of vertex positions after smoothing
    skel_verts_orig: np.array
        a N long index of skeleton vertices in the original mesh vertex index
    (mesh_to_skeleton_map): np.array
        a Mx2 map of mesh vertex indices to skeleton vertex indices

    """

    skeletonize_output = skeletonize_components(
        mesh,
        soma_pt=soma_pt,
        soma_thresh=soma_thresh,
        invalidation_d=invalidation_d,
        cc_vertex_thresh=cc_vertex_thresh,
        return_map=return_map,
        root_index=root_index,
    )
    if return_map is True:
        all_paths, roots, tot_path_lengths, mesh_to_skeleton_map = skeletonize_output
    else:
        all_paths, roots, tot_path_lengths = skeletonize_output

    all_edges = []
    for comp_paths in all_paths:
        all_edges.append(utils.paths_to_edges(comp_paths))
    if len(all_edges) > 0:
        tot_edges = np.vstack(all_edges)
    else:
        tot_edges = np.zeros((3, 0))

    skel_verts, skel_edges, skel_verts_orig = reduce_verts(mesh.vertices, tot_edges)

    if return_map:
        mesh_to_skeleton_map = utils.nanfilter_shapes(
            np.unique(tot_edges.ravel()), mesh_to_skeleton_map
        )
        mesh_to_skeleton_map[np.isnan(mesh_to_skeleton_map)] = -1
    else:
        mesh_to_skeleton_map = None

    output_tuple = (skel_verts, skel_edges, skel_verts_orig)

    if return_map:
        output_tuple = output_tuple + (mesh_to_skeleton_map.astype(int),)

    return output_tuple


def reduce_verts(verts, faces):
    """removes unused vertices from a graph or mesh

    Parameters
    ----------
    verts : numpy.array
        NxD numpy array of vertex locations
    faces : numpy.array
        MxK numpy array of connected shapes (i.e. edges or tris)
        (entries are indices into verts)

    Returns
    -------
    np.array
        new_verts, a filtered set of vertices
    np.array
        new_face, a reindexed set of faces (or edges)
    np.array
        used_verts, the index of the new_verts in the old verts

    """
    used_verts = np.unique(faces.ravel())
    new_verts = verts[used_verts, :]
    new_face = np.zeros(faces.shape, dtype=faces.dtype)
    for i in range(faces.shape[1]):
        new_face[:, i] = np.searchsorted(used_verts, faces[:, i])
    return new_verts, new_face, used_verts


def skeletonize_components(
    mesh,
    soma_pt=None,
    soma_thresh=10000,
    invalidation_d=10000,
    cc_vertex_thresh=100,
    return_map=False,
    root_index=None,
):
    """core skeletonization routine, used by :func:`meshparty.skeletonize.calculate_skeleton_paths_on_mesh`
    to calculate skeleton on all components of mesh, with no post processing"""
    # find all the connected components in the mesh
    n_components, labels = sparse.csgraph.connected_components(
        mesh.csgraph, directed=False, return_labels=True
    )
    comp_labels, comp_counts = np.unique(labels, return_counts=True)

    if return_map:
        mesh_to_skeleton_map = np.full(len(mesh.vertices), np.nan)

    # variables to collect the paths, roots and path lengths
    all_paths = []
    roots = []
    tot_path_lengths = []

    if root_index is not None:
        soma_d = np.linalg.norm(mesh.vertices - mesh.vertices[root_index], axis=1)
        is_soma_pt = np.arange(len(mesh.vertices)) == root_index
    elif soma_pt is not None:
        soma_d = mesh.vertices - soma_pt.reshape(1, 3)
        soma_d = np.linalg.norm(soma_d, axis=1)
        is_soma_pt = soma_d < soma_thresh
    else:
        is_soma_pt = None
        soma_d = None
    # is_soma_pt = None
    # soma_d = None

    # loop over the components
    for k in range(n_components):
        if comp_counts[k] > cc_vertex_thresh:

            # find the root using a soma position if you have it
            # it will fall back to a heuristic if the soma
            # is too far away for this component
            root, root_ds, pred, valid = setup_root(
                mesh, is_soma_pt, soma_d, labels == k
            )
            # run teasar on this component
            teasar_output = mesh_teasar(
                mesh,
                root=root,
                root_ds=root_ds,
                root_pred=pred,
                valid=valid,
                invalidation_d=invalidation_d,
                return_map=return_map,
            )
            if return_map is False:
                paths, path_lengths = teasar_output
            else:
                paths, path_lengths, mesh_to_skeleton_map_single = teasar_output
                mesh_to_skeleton_map[
                    ~np.isnan(mesh_to_skeleton_map_single)
                ] = mesh_to_skeleton_map_single[~np.isnan(mesh_to_skeleton_map_single)]

            if len(path_lengths) > 0:
                # collect the results in lists
                tot_path_lengths.append(path_lengths)
                all_paths.append(paths)
                roots.append(root)

    if return_map:
        return all_paths, roots, tot_path_lengths, mesh_to_skeleton_map
    else:
        return all_paths, roots, tot_path_lengths


def setup_root(mesh, is_soma_pt=None, soma_d=None, is_valid=None):
    """ function to find the root index to use for this mesh """
    if is_valid is not None:
        valid = np.copy(is_valid)
    else:
        valid = np.ones(len(mesh.vertices), np.bool)
    assert len(valid) == mesh.vertices.shape[0]

    root = None
    # soma mode
    if is_soma_pt is not None:
        # pick the first soma as root
        assert len(soma_d) == mesh.vertices.shape[0]
        assert len(is_soma_pt) == mesh.vertices.shape[0]
        is_valid_root = is_soma_pt & valid
        valid_root_inds = np.where(is_valid_root)[0]
        if len(valid_root_inds) > 0:
            min_valid_root = np.nanargmin(soma_d[valid_root_inds])
            root = valid_root_inds[min_valid_root]
            root_ds, pred = sparse.csgraph.dijkstra(
                mesh.csgraph, directed=False, indices=root, return_predecessors=True
            )
        else:
            start_ind = np.where(valid)[0][0]
            root, target, pred, dm, root_ds = utils.find_far_points(
                mesh, start_ind=start_ind
            )
        valid[is_soma_pt] = False

    if root is None:
        # there is no soma close, so use far point heuristic
        start_ind = np.where(valid)[0][0]
        root, target, pred, dm, root_ds = utils.find_far_points(
            mesh, start_ind=start_ind
        )
    valid[root] = False
    assert np.all(~np.isinf(root_ds[valid]))
    return root, root_ds, pred, valid


def mesh_teasar(
    mesh,
    root=None,
    valid=None,
    root_ds=None,
    root_pred=None,
    soma_pt=None,
    soma_thresh=7500,
    invalidation_d=10000,
    return_timing=False,
    return_map=False,
    exclude_edges_sigma=None,
):
    """core skeletonization function used to skeletonize a single component of a mesh"""
    # if no root passed, then calculation one
    if root is None:
        root, root_ds, root_pred, valid = setup_root(
            mesh, soma_pt=soma_pt, soma_thresh=soma_thresh
        )
    # if root_ds have not be precalculated do so
    if root_ds is None:
        root_ds, root_pred = sparse.csgraph.dijkstra(
            mesh.csgraph, False, root, return_predecessors=True
        )
    # if certain vertices haven't been pre-invalidated start with just
    # the root vertex invalidated
    if valid is None:
        valid = np.ones(len(mesh.vertices), np.bool)
        valid[root] = False
    else:
        if len(valid) != len(mesh.vertices):
            raise Exception("valid must be length of vertices")

    if return_map == True:
        mesh_to_skeleton_dist = np.full(len(mesh.vertices), np.inf)
        mesh_to_skeleton_map = np.full(len(mesh.vertices), np.nan)

    total_to_visit = np.sum(valid)
    if np.sum(np.isinf(root_ds) & valid) != 0:
        print(np.where(np.isinf(root_ds) & valid))
        raise Exception("all valid vertices should be reachable from root")

    # vector to store each branch result
    paths = []

    # vector to store each path's total length
    path_lengths = []

    # keep track of the nodes that have been visited
    visited_nodes = [root]

    # counter to track how many branches have been counted
    k = 0

    # arrays to track timing
    start = time.time()
    time_arrays = [[], [], [], [], []]

    with tqdm(total=total_to_visit) as pbar:
        # keep looping till all vertices have been invalidated
        while np.sum(valid) > 0:
            k += 1
            t = time.time()
            # find the next target, farthest vertex from root
            # that has not been invalidated
            target = np.nanargmax(root_ds * valid)
            if np.isinf(root_ds[target]):
                raise Exception("target cannot be reached")
            time_arrays[0].append(time.time() - t)

            t = time.time()
            # figure out the longest this branch could be
            # by following the route from target to the root
            # and finding the first already visited node (max_branch)
            # The dist(root->target) - dist(root->max_branch)
            # is the maximum distance the shortest route to a branch
            # point from the target could possibly be,
            # use this bound to reduce the djisktra search radius for this target
            max_branch = target
            while max_branch not in visited_nodes:
                max_branch = root_pred[max_branch]
            max_path_length = root_ds[target] - root_ds[max_branch]

            # calculate the shortest path to that vertex
            # from all other vertices
            # up till the distance to the root
            ds, pred_t = sparse.csgraph.dijkstra(
                mesh.csgraph,
                False,
                target,
                limit=max_path_length,
                return_predecessors=True,
            )

            # pick out the vertex that has already been visited
            # which has the shortest path to target
            min_node = np.argmin(ds[visited_nodes])
            # reindex to get its absolute index
            branch = visited_nodes[min_node]
            # this is in the index of the point on the skeleton
            # we want this branch to connect to
            time_arrays[1].append(time.time() - t)

            t = time.time()
            # get the path from the target to branch point
            path = utils.get_path(target, branch, pred_t)
            visited_nodes += path[0:-1]
            # record its length
            assert ~np.isinf(ds[branch])
            path_lengths.append(ds[branch])
            # record the path
            paths.append(path)
            time_arrays[2].append(time.time() - t)

            t = time.time()
            # get the distance to all points along the new path
            # within the invalidation distance
            dm, _, sources = sparse.csgraph.dijkstra(
                mesh.csgraph,
                False,
                path,
                limit=invalidation_d,
                min_only=True,
                return_predecessors=True,
            )
            time_arrays[3].append(time.time() - t)

            t = time.time()
            # all such non infinite distances are within the invalidation
            # zone and should be marked invalid
            nodes_to_update = ~np.isinf(dm)
            marked = np.sum(valid & ~np.isinf(dm))
            if return_map == True:
                new_sources_closer = (
                    dm[nodes_to_update] < mesh_to_skeleton_dist[nodes_to_update]
                )
                mesh_to_skeleton_map[nodes_to_update] = np.where(
                    new_sources_closer,
                    sources[nodes_to_update],
                    mesh_to_skeleton_map[nodes_to_update],
                )
                mesh_to_skeleton_dist[nodes_to_update] = np.where(
                    new_sources_closer,
                    dm[nodes_to_update],
                    mesh_to_skeleton_dist[nodes_to_update],
                )

            valid[~np.isinf(dm)] = False

            # print out how many vertices are still valid
            pbar.update(marked)
            time_arrays[4].append(time.time() - t)
    # record the total time
    dt = time.time() - start

    out_tuple = (paths, path_lengths)
    if return_map:
        out_tuple = out_tuple + (mesh_to_skeleton_map,)
    if return_timing:
        out_tuple = out_tuple + (time_arrays, dt)

    return out_tuple


def smooth_graph(values, edges, mask=None, neighborhood=2, iterations=100, r=0.1):
    """smooths a spatial graph via iterative local averaging
    calculates the average value of neighboring values
    and relaxes the values toward that average

    Parameters
    ----------
    values : numpy.array
        a NxK numpy array of values, for example xyz positions
    edges : numpy.array
        a Mx2 numpy array of indices into values that are edges
    mask : numpy.array
        NOT yet implemented
        optional N boolean vector of values to mask
        the vert locations.  the result will return a result at every vert
        but the values that are false in this mask will be ignored and not
        factored into the smoothing.
    neighborhood : int
        an integer of how far in the graph to relax over
        as being local to any vertex (default = 2)
    iterations : int
        number of relaxation iterations (default = 100)
    r : float
        relaxation factor at each iteration
        new_vertex = (1-r)*old_vertex*mask + (r+(1-r)*(1-mask))*(local_avg)
        default = .1

    Returns
    -------
    np.array
        new_verts, a Nx3 list of new smoothed vertex positions

    """
    N = len(values)
    E = len(edges)

    # setup a sparse matrix with the edges
    sm = sparse.csc_matrix((np.ones(E), (edges[:, 0], edges[:, 1])), shape=(N, N))

    # an identity matrix
    eye = sparse.csc_matrix(
        (np.ones(N, dtype=np.float32), (np.arange(0, N), np.arange(0, N))), shape=(N, N)
    )
    # for undirected graphs we want it symettric
    sm = sm + sm.T

    # this will store our relaxation matrix
    C = sparse.csc_matrix(eye)
    # multiple the matrix and add to itself
    # to spread connectivity along the graph
    for i in range(neighborhood):
        C = C + sm @ C
    # zero out the diagonal elements
    C.setdiag(np.zeros(N))
    # don't overweight things that are connected in more than one way
    C = C.sign()
    # measure total effective neighbors per node
    neighbors = np.sum(C, axis=1)

    # normalize the weights of neighbors according to number of neighbors
    neighbors = 1 / neighbors
    C = C.multiply(neighbors)
    # convert back to csc
    C = C.tocsc()

    # multiply weights by relaxation term
    C *= r

    # construct relaxation matrix, adding on identity with complementary weight
    A = C + (1 - r) * eye

    # make a copy of original vertices to no destroy inpuyt
    new_values = np.copy(values)

    # iteratively relax the vertices
    for i in range(iterations):
        new_values = A * new_values
    return new_values


def soma_via_sphere(soma_pt, verts, edges, soma_d_thresh):
    """Get indices within soma_d_thresh of a soma_pt. Exclude vertices that left and come back."""
    closest_soma_ind = np.argmin(np.linalg.norm(verts - soma_pt, axis=1))
    close_inds = np.linalg.norm(verts - soma_pt, axis=1) < soma_d_thresh
    orig_graph = utils.create_csgraph(verts, edges, euclidean_weight=False)
    speye = sparse.diags(close_inds.astype(int))
    _, compids = sparse.csgraph.connected_components(orig_graph * speye)
    return np.flatnonzero(compids[closest_soma_ind] == compids), soma_d_thresh


def soma_via_branch_starts(
    sk,
    mesh,
    soma_pt,
    rs,
    search_radius=20000,
    fallback_radius=15000,
    cutoff_threshold=0.2,
    min_cutoff=0.1,
    dynamic_range=1,
    dynamic_threshold=False,
):
    """Runs down paths into the soma region and finds onset of each branch."""

    is_close = np.linalg.norm(sk.vertices - soma_pt, axis=1) < search_radius
    is_close_fallback = np.linalg.norm(sk.vertices - soma_pt, axis=1) < fallback_radius

    # Find segments that emerge from the soma region
    close_segs = []
    for seg in sk.segments:
        seg = seg[np.argsort(sk.distance_to_root[seg])]
        if is_close[seg[0]]:
            if np.all(is_close[sk.path_to_root(seg[0])]):
                close_segs.append(seg)
    close_seg_inds = np.concatenate(close_segs)
    close_inds = close_seg_inds[is_close[close_seg_inds]]

    # From those segments, find the tips that come out of the search_radius
    is_close_specific = np.full(sk.n_vertices, False)
    is_close_specific[close_inds] = True
    close_parent_edges = sk.edges[is_close_specific[sk.edges[:, 1]]]
    tip_inds = close_parent_edges[~is_close_specific[close_parent_edges[:, 0]], 0]

    rs = rs[close_inds]
    rs_long = np.full(sk.n_vertices, np.nan)
    rs_long[close_inds] = rs
    sk.reroot(close_inds[np.argmin(np.abs(rs - np.percentile(rs, 98)))])

    # Fit a logistic curve
    def log_func(x, h, k, xh, a):
        return h / (1 + np.exp(-k * (xh - x))) + a

    all_params = []
    soma_votes = []

    for tip_ind in tip_inds:
        ptr = sk.path_to_root(tip_ind)
        path_inds = ptr[1:]
        xdata = sk.distance_to_root[path_inds] / 1000
        ydata = rs_long[path_inds] / 1000
        good_rows = np.invert(np.logical_or(np.isnan(ydata), np.isinf(ydata)))
        ydata = ydata[good_rows]
        xdata = xdata[good_rows]
        ydata_filt = np.maximum.accumulate(signal.medfilt(ydata, 21))

        try:
            # sig = ydata_filt * log_func( (np.max(xdata)-xdata), 1, 2, 3, 1)
            sig = np.where(ydata_filt < 2, 2, ydata_filt) * log_func(
                (np.max(xdata) - xdata), 2, 1, 5, 1
            )
            params, _ = optimize.curve_fit(
                log_func,
                xdata,
                ydata_filt,
                sigma=sig,
                bounds=([0, 0.5, 0, 0], [np.inf, 5, np.inf, np.inf]),
                p0=(10, 1, 10, 1),
                method="trf",
            )
            all_params.append(params)
            if dynamic_threshold:
                cutoff_threshold_eff = np.min(
                    [
                        cutoff_threshold,
                        min_cutoff
                        + (cutoff_threshold - min_cutoff)
                        * np.max([0, (params[1] - 0.5) / dynamic_range]),
                    ]
                )
            else:
                cutoff_threshold_eff = cutoff_threshold

            def f(x):
                return log_func(x, *params) - (
                    params[3] + cutoff_threshold_eff * (params[0] - params[3])
                )

            opt_sol = optimize.root_scalar(f, bracket=[0, xdata.max()])
            if opt_sol.converged:
                root = opt_sol.root
                use_fallback = False
            else:
                use_fallback = True
        except:
            use_fallback = True
            all_params.append(None)

        if not use_fallback:
            base_path_ind = np.argmin(np.abs(xdata - root))
        else:
            d_to_soma_pt = np.linalg.norm(sk.vertices[path_inds] - soma_pt, axis=1)
            base_path_ind = np.argmin(np.abs(d_to_soma_pt - fallback_radius))

        soma_vote = np.full(sk.n_vertices, np.nan)
        soma_vote[path_inds[base_path_ind:]] = 1
        soma_vote[path_inds[:base_path_ind]] = 0
        soma_votes.append(soma_vote)

    # Any tips whose path to root is entirely in the close zone also vote as as 'somatic'
    for ep in sk.end_points[is_close[sk.end_points]]:
        ptr = sk.path_to_root(ep)
        if np.all(is_close[ptr]):
            soma_vote = np.full(sk.n_vertices, np.nan)
            soma_vote[ptr] = 1
            soma_vote[ptr[is_close_fallback[ptr]]] = np.inf
            soma_votes.append(soma_vote)

    # Get soma region
    soma_votes = np.vstack(soma_votes)
    num_votes = len(soma_votes) - np.sum(np.isnan(soma_votes), axis=0)
    num_yes = np.nansum(soma_votes, axis=0)

    with np.errstate(all="ignore"):
        is_soma = (num_yes / num_votes) > 0.5

    last_nonsoma = []
    for tip_ind in tip_inds:
        ptr = sk.path_to_root(tip_ind)
        last_nonsoma.append(ptr[np.flatnonzero(np.diff(is_soma[ptr]) == 1)[0]])
    last_nonsoma = np.unique(last_nonsoma)

    keep_binds = []
    for bind in last_nonsoma:
        bind_ptr = sk.path_to_root(bind)
        if np.any(np.isin(last_nonsoma, bind_ptr[1:])):
            keep_binds.append(False)
        else:
            keep_binds.append(True)
    keep_binds = np.array(keep_binds)

    g = sk.cut_graph(last_nonsoma[keep_binds])
    _, comps = sparse.csgraph.connected_components(g)
    root_comp = comps[sk.root]
    return np.flatnonzero(comps == root_comp), np.nanmedian(rs_long[comps == root_comp])


def collapse_soma_skeleton(
    soma_verts,
    soma_pt,
    verts,
    edges,
    mesh_to_skeleton_map=None,
    collapse_index=None,
    return_filter=False,
    return_soma_ind=False,
):
    """function to adjust skeleton result to move root to soma_pt

    Parameters
    ----------
    soma_pt : numpy.array
        a 3 long vector of xyz locations of the soma (None to just remove duplicate )
    verts : numpy.array
        a Nx3 array of xyz vertex locations
    edges : numpy.array
        a Kx2 array of edges of the skeleton
    soma_d_thresh : float
        distance from soma_pt to collapse skeleton nodes
    mesh_to_skeleton_map : np.array
        a M long array of how each mesh index maps to a skeleton vertex
        (default None).  The function will update this as it collapses vertices to root.
    soma_mesh_indices : np.array
         a K long array of indices in the mesh that should be considered soma
         Any  skeleton vertex on these vertices will all be collapsed to root.
    return_filter : bool
        whether to return a list of which skeleton vertices were used in the end
        for the reduced set of skeleton vertices
    only_soma_component : bool
        whether to collapse only the skeleton connected component which is closest to the soma_pt
        (default True)
    return_soma_ind : bool
        whether to return which skeleton index that is the soma_pt

    Returns
    -------
    np.array
        verts, Px3 array of xyz skeleton vertices
    np.array
        edges, Qx2 array of skeleton edges
    (np.array)
        new_mesh_to_skeleton_map, returned if mesh_to_skeleton_map and soma_pt passed
    (np.array)
        used_vertices, if return_filter this contains the indices into the passed verts which the return verts is using
    int
        an index into the returned verts that is the root of the skeleton node, only returned if return_soma_ind is True

    """
    if soma_verts is not None:
        soma_pt_m = soma_pt.reshape(1, 3)
        if collapse_index is None:
            new_verts = np.vstack((verts, soma_pt_m))
            soma_i = verts.shape[0]
        else:
            new_verts = verts
            soma_i = collapse_index
            soma_verts = soma_verts[soma_verts != soma_i]

        edges_m = edges.copy()
        edges_m[np.isin(edges, soma_verts)] = soma_i

        simple_verts, simple_edges = utils.remove_unused_verts(new_verts, edges_m)
        good_edges = ~(simple_edges[:, 0] == simple_edges[:, 1])

        if mesh_to_skeleton_map is not None:
            consolidate_dict = {v: soma_i for v in soma_verts}
            new_index_dict, _ = utils.remap_dict(len(new_verts), consolidate_dict)
            new_index_dict[-1] = -1
            mesh_to_skeleton_map[np.isnan(mesh_to_skeleton_map)] = -1
            new_mesh_to_skeleton_map = fastremap.remap(
                mesh_to_skeleton_map, new_index_dict
            )

        output = [simple_verts, simple_edges[good_edges]]
        if mesh_to_skeleton_map is not None:
            output.append(new_mesh_to_skeleton_map)
        if return_filter:
            used_vertices = np.unique(edges_m.ravel())
            if collapse_index is None:
                # Remove the largest value which is soma_i
                used_vertices = used_vertices[:-1]
            output.append(used_vertices)
        if return_soma_ind:
            output.append(new_index_dict[soma_i])
        return output

    else:
        simple_verts, simple_edges = utils.remove_unused_verts(verts, edges)
        return simple_verts, simple_edges


def resample(sk, spacing, kind="linear", tip_length_ratio=0.25):
    """Resample a skeleton's vertices

    Parameters
    ----------
    sk : Skeleton
        Input skeleton file with a skeleton
    spacing : numeric
        Desired spacing in nanometers
    kind : str, optional
        Type of interpolation to use when resampling. Options follow scipy.interpolate.interp1d. By default "linear"
    tip_length_ratio : float, optional
        The ratio of spacing to branch tip length that a branch tip must have in order to be included in the final skeleton
        for example: spacing is 10 and branch length is 8. do you want to include that final 8 length tip?
        then perhaps consider a tip_length_ratio of .75, by default 0.25

    Returns
    -------
    Skeleton
        New skeleton with resampled vertices.

    resample_map
        Array where the ith index corresponds to the ith vertex of the resampled skeleton and the value
        is the associated index in the original skeleton. To assign vertices, we assign a "domain" to each
        vertex in the original skeleton that is halfway between the vertex and its neighbors. Resampled
        vertices that fall within that domain (based on topology and distance-to-root) are then associated
        with the original vertex.
    """
    path_counter = 0
    branch_d = {}
    vert_list = []
    edge_list = []
    output_map_list = []

    for path in sk.cover_paths:
        new_verts, new_edges, output_map_path, branch_d = skeleton_utils.resample_path(
            path,
            sk,
            path_counter,
            spacing,
            kind,
            tip_length_ratio,
            branch_d,
        )
        vert_list.append(new_verts)
        edge_list.append(new_edges)
        output_map_list.append(output_map_path)
        path_counter += len(new_verts)

    new_verts = np.vstack(vert_list)
    new_edges = np.vstack(edge_list)
    resample_map = np.concatenate(output_map_list)

    return (
        Skeleton(
            new_verts,
            new_edges,
            root=branch_d[int(sk.root)],
            remove_zero_length_edges=False,
        ),
        resample_map,
    )


import numpy as np
from meshparty import utils
from scipy import spatial, sparse
from dataclasses import dataclass, fields, asdict, make_dataclass

try:
    from pykdtree.kdtree import KDTree as pyKDTree
except:
    pyKDTree = spatial.cKDTree
from meshparty import skeleton_io
from collections.abc import Iterable
from .skeleton_utils import resample_path


def _metadata_from_dict(
    meta_dict,
    dataclass_name="MetaMetadata",
):
    meta = make_dataclass(dataclass_name, fields=meta_dict.keys())
    return meta(**meta_dict)


@dataclass
class SkeletonMetadata:
    root_id: int = None
    soma_pt_x: float = None
    soma_pt_y: float = None
    soma_pt_z: float = None
    soma_radius: float = None
    collapse_soma: bool = None
    collapse_function: str = None
    invalidation_d: float = None
    smooth_vertices: bool = None
    compute_radius: bool = None
    shape_function: str = None
    smooth_iterations: int = None
    smooth_neighborhood: int = None
    smooth_r: float = None
    cc_vertex_thresh: int = None
    remove_zero_length_edges: bool = None
    collapse_params: dict = None
    timestamp: float = None
    skeleton_type: str = None
    meta: object = None

    # Fields used for skeletonization
    _skeletonize_fields = [
        "soma_pt",
        "soma_radius",
        "collapse_soma",
        "collapse_function",
        "invalidation_d",
        "smooth_vertices",
        "compute_radius",
        "shape_function",
        "smooth_iterations",
        "smooth_neighborhood",
        "smooth_r",
        "cc_vertex_thresh",
        "remove_zero_length_edges",
        "collapse_params",
    ]

    def __init__(self, **kwargs):
        names = [f.name for f in fields(self)]

        if kwargs.get("meta") is not None:
            setattr(
                self, "meta", _metadata_from_dict(kwargs.pop("meta"), "MetaMetadata")
            )

        for k, v in kwargs.items():
            if k in names:
                if isinstance(v, np.ndarray):
                    v = v.tolist()
                setattr(self, k, v)

    def skeletonize_kwargs(self):
        params = asdict(self)

        # reassemble soma point into list
        soma_pt = [
            params.pop("soma_pt_x"),
            params.pop("soma_pt_y"),
            params.pop("soma_pt_z"),
        ]
        if soma_pt[0] is not None:
            params["soma_pt"] = soma_pt
        else:
            params["soma_pt"] = None

        for k in list(params.keys()):
            if k not in self._skeletonize_fields:
                params.pop(k)
        return params

    def update_metameta(self, metameta):
        if self.meta is not None:
            meta_dict = asdict(self.meta)
        else:
            meta_dict = {}

        meta_dict.update(metameta)
        setattr(self, "meta", _metadata_from_dict(meta_dict, "MetaMetadata"))
        pass


class StaticSkeleton:
    def __init__(
        self,
        vertices,
        edges,
        root=None,
        radius=None,
        mesh_to_skel_map=None,
        mesh_index=None,
        vertex_properties=None,
        voxel_scaling=None,
    ):
        self._vertices = vertices
        self._edges = edges
        self._root = None
        self._radius = radius
        self._mesh_to_skel_map = mesh_to_skel_map
        self._mesh_index = mesh_index

        self._parent_node_array = None
        self._distance_to_root = None
        self._csgraph = None
        self._csgraph_binary = None
        self._voxel_scaling = voxel_scaling

        if root is None:
            self._create_default_root()
        else:
            self.reroot(root, reset_other_components=True)

        self._reset_derived_properties()
        self.vertex_properties = vertex_properties

    @property
    def vertices(self):
        return self._vertices

    @property
    def edges(self):
        return self._edges

    @property
    def mesh_to_skel_map(self):
        return self._mesh_to_skel_map

    @property
    def voxel_scaling(self):
        if self._voxel_scaling is None:
            return None
        else:
            return np.array(self._voxel_scaling)

    @voxel_scaling.setter
    def voxel_scaling(self, new_scaling):
        self._vertices = self._vertices * self.inverse_voxel_scaling
        if new_scaling is not None:
            self._vertices = self._vertices * np.array(new_scaling).reshape(3)
        self._voxel_scaling = new_scaling
        self._reset_derived_properties()

    @property
    def inverse_voxel_scaling(self):
        if self.voxel_scaling is None:
            return np.array([1, 1, 1])
        else:
            return 1 / self.voxel_scaling

    @property
    def n_vertices(self):
        """ int : Number of vertices in the skeleton """
        return len(self.vertices)

    @property
    def root(self):
        """ int : Index of the skeleton root """
        if self._root is None:
            self._create_default_root()
        return self._root

    @property
    def radius(self):
        if self._radius is None:
            return None
        else:
            return self._radius

    @radius.setter
    def radius(self, new_values):
        if len(new_values) == self.n_vertices:
            self._radius = np.array(new_values).reshape(self.n_vertices)

    @property
    def mesh_index(self):
        return self._mesh_index

    def _create_default_root(self):
        temp_graph = utils.create_csgraph(
            self.vertices, self.edges, euclidean_weight=True, directed=False
        )
        r = utils.find_far_points_graph(temp_graph)
        self.reroot(int(r[0]), reset_other_components=True)

    def reroot(self, new_root, reset_other_components=False):
        """Change the skeleton root index.

        Parameters
        ----------
        new_root : Int
            Skeleton vertex index to be the new root.

        reset_other_components : Bool
            Orders non-root components accoring to a local default "root".
            Should not often be set to True by a user.
        """
        if new_root > self.n_vertices:
            raise ValueError("New root must correspond to a skeleton vertex index")
        self._root = int(new_root)
        self._parent_node_array = np.full(self.n_vertices, None)

        _, lbls = sparse.csgraph.connected_components(self.csgraph_binary)
        root_comp = lbls[new_root]
        if reset_other_components:
            comps_to_reroot = np.unique(lbls)
        else:
            comps_to_reroot = [root_comp]
        # The edge list has to be treated like an undirected graph

        for comp in comps_to_reroot:
            if comp == root_comp:
                comp_root = new_root
            else:
                comp_root = utils.find_far_points_graph(
                    self.csgraph_binary,
                    start_ind=np.flatnonzero(lbls == comp)[0],
                    multicomponent=True,
                )[0]

            d = sparse.csgraph.dijkstra(
                self.csgraph_binary, directed=False, indices=comp_root
            )

            # Make edges in edge list orient as [child, parent]
            # Where each child only has one parent
            # And the root has no parent. (Thus parent is closer than child)
            edge_slice = np.any(
                np.isin(self.edges, np.flatnonzero(lbls == comp)), axis=1
            )

            edges = self.edges[edge_slice]
            is_ordered = d[edges[:, 0]] > d[edges[:, 1]]
            e1 = np.where(is_ordered, edges[:, 0], edges[:, 1])
            e2 = np.where(is_ordered, edges[:, 1], edges[:, 0])
            self._edges[edge_slice] = np.stack((e1, e2)).T
            self._parent_node_array[e1] = e2

        self._reset_derived_properties()

    ######################
    # Derived properties #
    ######################

    def _reset_derived_properties(self):
        self._csgraph = None
        self._csgraph_binary = None
        self._distance_to_root = None

    @property
    def csgraph(self):
        if self._csgraph is None:
            self._csgraph = utils.create_csgraph(
                self.vertices, self.edges, euclidean_weight=True, directed=True
            )
        return self._csgraph

    @property
    def csgraph_binary(self):
        if self._csgraph_binary is None:
            self._csgraph_binary = utils.create_csgraph(
                self.vertices, self.edges, euclidean_weight=False, directed=True
            )
        return self._csgraph_binary

    @property
    def csgraph_undirected(self):
        return self.csgraph + self.csgraph.T

    @property
    def csgraph_binary_undirected(self):
        return self.csgraph_binary + self.csgraph_binary.T

    def parent_nodes(self, vinds):
        """Get a list of parent nodes for specified vertices

        Parameters
        ----------
        vinds : Collection of ints
            Collection of vertex indices

        Returns
        -------
        numpy.array
            The parent node of each vertex index in vinds.
        """
        if isinstance(vinds, list):
            vinds = np.array(vinds)
        return self._parent_node_array[vinds]

    @property
    def distance_to_root(self):
        """np.array : N length array with the distance to the root node along the skeleton."""
        if self._distance_to_root is None:
            self._distance_to_root = sparse.csgraph.dijkstra(
                self.csgraph, directed=False, indices=self.root
            )
        return self._distance_to_root

    def path_to_root(self, v_ind):
        """
        Gives the path to root from a specified vertex.

        Parameters
        ----------
        v_ind : int
            Vertex index

        Returns
        -------
        numpy.array : Ordered set of indices from v_ind to root, inclusive of both.
        """
        path = [v_ind]
        ind = v_ind
        ind = self._parent_node_array[ind]
        while ind is not None:
            path.append(ind)
            ind = self._parent_node_array[ind]
        return np.array(path)


class Skeleton:
    def __init__(
        self,
        vertices,
        edges,
        root=None,
        radius=None,
        mesh_to_skel_map=None,
        mesh_index=None,
        vertex_properties={},
        node_mask=None,
        voxel_scaling=None,
        remove_zero_length_edges=True,
        skeleton_index=None,
        meta={},
    ):

        if remove_zero_length_edges:
            zlsk = utils.collapse_zero_length_edges(
                vertices,
                edges,
                root,
                radius,
                mesh_to_skel_map,
                mesh_index,
                node_mask,
                vertex_properties,
            )
            (
                vertices,
                edges,
                root,
                radius,
                mesh_to_skel_map,
                mesh_index,
                node_mask,
                vertex_properties,
            ) = zlsk

        self._rooted = StaticSkeleton(
            vertices,
            edges,
            radius=radius,
            mesh_to_skel_map=mesh_to_skel_map,
            mesh_index=mesh_index,
            vertex_properties=vertex_properties,
            root=root,
            voxel_scaling=voxel_scaling,
        )

        self._node_mask = np.full(self._rooted.n_vertices, True)
        self._edges = None
        self._SkeletonIndex = skeleton_index

        # Derived properties of the filtered graph
        self._csgraph_filtered = None
        self._cover_paths = None
        self._segments = None
        self._segment_map = None
        self._kdtree = None
        self._pykdtree = None
        self._reset_derived_properties_filtered()
        self.vertex_properties = vertex_properties

        if isinstance(meta, SkeletonMetadata):
            self._meta = meta
        else:
            self._meta = SkeletonMetadata(**meta)
        if node_mask is not None:
            self.apply_mask(node_mask, in_place=True)

    @property
    def meta(self):
        return self._meta

    ###################
    # Mask properties #
    ###################

    @property
    def SkeletonIndex(self):
        if self._SkeletonIndex is None:
            self._SkeletonIndex = np.array
        return self._SkeletonIndex

    def _register_skeleton_index(self, NewSkeletonIndex):
        self._SkeletonIndex = NewSkeletonIndex

    @property
    def node_mask(self):
        return self._node_mask

    def copy(self):
        return Skeleton(
            self._rooted.vertices,
            self._rooted.edges,
            mesh_to_skel_map=self._rooted.mesh_to_skel_map,
            vertex_properties=self._rooted.vertex_properties,
            root=self._rooted.root,
            node_mask=self.node_mask,
            radius=self._rooted.radius,
            voxel_scaling=self.voxel_scaling,
            skeleton_index=self._SkeletonIndex,
            mesh_index=self._rooted.mesh_index,
            remove_zero_length_edges=False,
            meta=self.meta,
        )

    def apply_mask(self, new_mask, in_place=False):
        if in_place:
            sk = self
        else:
            sk = self.copy()

        if len(new_mask) == len(sk.vertices):
            all_false = np.full(len(sk.node_mask), False)
            all_false[sk.node_mask] = new_mask
            new_mask = all_false

        sk._node_mask = new_mask
        sk._reset_derived_properties_filtered()

        if in_place is False:
            return sk

    def reset_mask(self, in_place=False):
        true_mask = np.full(self.unmasked_size, True)
        out = self.apply_mask(true_mask, in_place=in_place)
        if in_place is False:
            return out

    def mask_from_indices(self, mask_indices):
        new_mask = np.full(self._rooted.n_vertices, False)
        new_mask[self.map_indices_to_unmasked(mask_indices)] = True
        return new_mask

    @property
    def indices_unmasked(self):
        """
        np.array: Gets the indices of nodes in the filtered mesh in the unmasked index array
        """
        return np.flatnonzero(self.node_mask)

    @property
    def unmasked_size(self):
        return len(self._rooted.vertices)

    def map_indices_to_unmasked(self, unmapped_indices):
        """
        For a set of masked indices, returns the corresponding unmasked indices

        Parameters
        ----------
        unmapped_indices: np.array
            a set of indices in the masked index space

        Returns
        -------
        np.array
            the indices mapped back to the original mesh index space
        """
        return utils.map_indices_to_unmasked(self.indices_unmasked, unmapped_indices)

    def map_boolean_to_unmasked(self, unmapped_boolean):
        """
        For a boolean index in the masked indices, returns the corresponding unmasked boolean index

        Parameters
        ----------
        unmapped_boolean : np.array
            a bool array in the masked index space

        Returns
        -------
        np.array
            a bool array in the original index space.  Is True if the unmapped_boolean suggests it should be.
        """
        return utils.map_boolean_to_unmasked(
            self.unmasked_size, self.node_mask, unmapped_boolean
        )

    def filter_unmasked_boolean(self, unmasked_boolean):
        """
        For an unmasked boolean slice, returns a boolean slice filtered to the masked mesh

        Parameters
        ----------
        unmasked_boolean : np.array
            a bool array in the original mesh index space

        Returns
        -------
        np.array
            returns the elements of unmasked_boolean that are still relevant in the masked index space
        """
        return utils.filter_unmasked_boolean(self.node_mask, unmasked_boolean)

    def filter_unmasked_indices(self, unmasked_shape, mask=None):
        """
        filters a set of indices in the original mesh space
        and returns it in the masked space

        Parameters
        ----------
        unmasked_shape: np.array
            a set of indices into vertices in the unmasked index space
        mask: np.array or None
            the mask to apply. default None will use this Mesh node_mask

        Returns
        -------
        np.array
            the unmasked_shape indices mapped into the masked index space
        """
        if mask is None:
            mask = self.node_mask
        return utils.filter_unmasked_indices(mask, unmasked_shape)

    def filter_unmasked_indices_padded(self, unmasked_shape, mask=None):
        """
        filters a set of indices in the original mesh space
        and returns it in the masked space

        Parameters
        ----------
        unmasked_shape: np.array
            a set of indices into vertices in the unmasked index space
        mask: np.array or None
            the mask to apply. default None will use this Mesh node_mask

        Returns
        -------
        np.array
            the unmasked_shape indices mapped into the masked index space,
            with -1 where the original index did not map into the masked mesh.
        """
        if mask is None:
            mask = self.node_mask
        return utils.filter_unmasked_indices_padded(mask, unmasked_shape)

    ####################
    # Basic properties #
    ####################

    @property
    def vertices(self):
        if self._vertices is None:
            self._vertices = self._rooted.vertices[self.node_mask]
        return self._vertices

    @vertices.setter
    def vertices(self, new_vertices):
        new_vertices = np.atleast_2d(new_vertices)
        if new_vertices.shape[1] != 3:
            raise ValueError("New vertices must be 3 dimensional")
        if len(new_vertices) == self._rooted.n_vertices:
            self._rooted._vertices = new_vertices
        elif len(new_vertices) == self.n_vertices:
            self._rooted._vertices[self.node_mask] = new_vertices
        else:
            raise ValueError("New vertices must be the same size as existing vertices")
        self._reset_derived_properties_rooted()
        self._reset_derived_properties_filtered(index_changed=False)

    @property
    def edges(self):
        if self._edges is None:
            self._edges = self.filter_unmasked_indices(self._rooted.edges)
        return self._edges

    @property
    def n_vertices(self):
        """ int : Number of vertices in the skeleton """
        return len(self.vertices)

    @property
    def mesh_to_skel_map(self):
        """numpy.array : N_mesh length array giving the associated skeleton
        vertex for each mesh vertex"""
        if self._rooted.mesh_to_skel_map is None:
            return None
        else:
            return self.filter_unmasked_indices_padded(self._rooted.mesh_to_skel_map)

    @property
    def mesh_to_skel_map_base(self):
        """numpy.array : N_mesh length array giving the associated skeleton
        vertex for each mesh vertex"""
        if self._rooted.mesh_to_skel_map is None:
            return None
        else:
            return self._rooted.mesh_to_skel_map

    @property
    def radius(self):
        if self._rooted.radius is None:
            return None
        return self._rooted.radius[self.node_mask]

    @property
    def mesh_index(self):
        if self._rooted.mesh_index is None:
            return None
        return self._rooted.mesh_index[self.node_mask]

    @property
    def csgraph(self):
        return self._rooted.csgraph[:, self.node_mask][self.node_mask]

    @property
    def csgraph_binary(self):
        return self._rooted.csgraph_binary[:, self.node_mask][self.node_mask]

    @property
    def csgraph_undirected(self):
        return self._rooted.csgraph_undirected[:, self.node_mask][self.node_mask]

    @property
    def csgraph_binary_undirected(self):
        return self._rooted.csgraph_binary_undirected[:, self.node_mask][self.node_mask]

    ##################
    # Voxel scalings #
    ##################

    @property
    def voxel_scaling(self):
        return self._rooted.voxel_scaling

    @voxel_scaling.setter
    def voxel_scaling(self, new_scaling):
        self._rooted.voxel_scaling = new_scaling
        self._reset_derived_properties_filtered(index_changed=False)

    #####################
    # Rooted properties #
    #####################

    def _reset_derived_properties_rooted(self):
        self._rooted._reset_derived_properties()

    def _create_default_root(self):
        temp_graph = utils.create_csgraph(
            self._rooted.vertices,
            self._rooted.edges,
            euclidean_weight=True,
            directed=False,
        )
        r = utils.find_far_points_graph(temp_graph)
        self._rooted.reroot(int(r[0]))

    @property
    def root(self):
        return self.SkeletonIndex(
            self.filter_unmasked_indices_padded(self._rooted.root)
        )

    @property
    def root_position(self):
        return self._rooted.vertices[self._rooted.root]

    def reroot(self, new_root):
        self._rooted.reroot(self.map_indices_to_unmasked(new_root))
        self._reset_derived_properties_filtered()

    @property
    def distance_to_root(self):
        "Distance to root (even if root is not in the mask)"
        return self._rooted.distance_to_root[self.node_mask]

    def path_to_root(self, v_ind):
        "Path stops if it leaves masked region"
        path_b = self._rooted.path_to_root(self.map_indices_to_unmasked(v_ind))
        path_filt = self.filter_unmasked_indices_padded(path_b)
        if np.any(path_filt == -1):
            last_ind = np.flatnonzero(path_filt == -1)[0]
        else:
            last_ind = len(path_filt)
        return self.SkeletonIndex(path_filt[:last_ind])

    #######################
    # Filtered properties #
    #######################

    def _reset_derived_properties_filtered(self, index_changed=True):
        self._vertices = None
        self._edges = None
        self._kdtree = None
        self._pykdtree = None

        if index_changed:
            self._branch_points = None
            self._end_points = None
            self._segment_map = None
            self._SkeletonIndex = None
            self._cover_paths = None

    #########################
    # Geometric quantitites #
    #########################

    @property
    def kdtree(self):
        """ scipy.spatial.kdtree : k-D tree from scipy.spatial. """
        if self._kdtree is None:
            self._kdtree = spatial.cKDTree(self.vertices)
        return self._kdtree

    @property
    def pykdtree(self):
        if self._pykdtree is None:
            self._pykdtree = pyKDTree(self.vertices)
        return self._pykdtree

    def _single_path_length(self, path):
        """Compute the length of a single path (assumed to be correct)"""
        path = np.unique(path)
        return np.sum(self.csgraph[:, path][path])

    def path_length(self, paths=None):
        """Returns the length of a path (described as an ordered collection of connected indices)
        Parameters
        ----------
        path : Path or collection of paths, a path being an ordered list of linked vertex indices.

        Returns
        -------
        Float or list of floats : The length of each path.
        """
        if paths is None:
            paths = np.arange(self.n_vertices)

        if len(paths) == 0:
            return 0

        if isinstance(paths[0], Iterable):
            Ls = []
            for path in paths:
                Ls.append(self._single_path_length(path))
        else:
            Ls = self._single_path_length(paths)
        return Ls

    ################################
    # Topological split properties #
    ################################

    def _create_branch_and_end_points(self):
        """Pre-compute branch and end points from the graph"""
        n_children = np.sum(self.csgraph_binary > 0, axis=0).squeeze()
        self._branch_points = np.flatnonzero(n_children > 1)
        self._end_points = np.flatnonzero(n_children == 0)

    @property
    def branch_points(self):
        """ numpy.array : Indices of branch points on the skeleton (pottentially including root)"""
        if self._branch_points is None:
            self._create_branch_and_end_points()
        return self.SkeletonIndex(self._branch_points)

    @property
    def end_points(self):
        """ numpy.array : Indices of end points on the skeleton (pottentially including root)"""
        if self._end_points is None:
            self._create_branch_and_end_points()
        return self.SkeletonIndex(self._end_points)

    @property
    def topo_points(self):
        return self.SkeletonIndex(
            np.concatenate([self.end_points, self.branch_points, [self.root]])
        )

    @property
    def n_branch_points(self):
        return len(self.branch_points)

    @property
    def n_end_points(self):
        return len(self.end_points)

    @property
    def end_points_undirected(self):
        """End points without skeleton orientation, including root and disconnected components."""
        return self.SkeletonIndex(
            np.flatnonzero(np.sum(self.csgraph_binary_undirected, axis=0) == 1)
        )

    @property
    def branch_points_undirected(self):
        return self.SkeletonIndex(
            np.flatnonzero(np.sum(self.csgraph_binary_undirected, axis=0) > 2)
        )

    def _compute_segments(self):
        """Precompute segments between branches and end points"""
        segments = []
        segment_map = np.zeros(len(self.vertices)) - 1

        if len(self.branch_points) > 0:
            ch_inds = np.concatenate(self.child_nodes(self.branch_points))
            g = self.cut_graph(ch_inds)
            _, ls = sparse.csgraph.connected_components(g)
        else:
            _, ls = sparse.csgraph.connected_components(self.csgraph_binary)

        _, invs = np.unique(ls, return_inverse=True)
        segments = [
            self.SkeletonIndex(np.flatnonzero(invs == ii)) for ii in np.unique(invs)
        ]
        segment_map = invs

        return segments, segment_map.astype(int)

    @property
    def segments(self):
        """list : A list of numpy.array indicies of segments, paths from each branch or
        end point (inclusive) to the next rootward branch/root point (exclusive), that
        cover the skeleton"""
        if self._segments is None:
            self._segments, self._segment_map = self._compute_segments()
        return self._segments

    @property
    def segment_map(self):
        """np.array : N set of of indices between 0 and len(self.segments)-1, denoting
        which segment a given skeleton vertex is in.
        """
        if self._segment_map is None:
            self._segments, self._segment_map = self._compute_segments()
        return self._segment_map

    def path_between(self, s_ind, t_ind):
        d, Ps = sparse.csgraph.dijkstra(
            self.csgraph_binary_undirected,
            directed=False,
            indices=s_ind,
            return_predecessors=True,
        )
        if not np.isinf(d[t_ind]):
            return self.SkeletonIndex(utils.path_from_predecessors(Ps, t_ind))
        else:
            return None

    ############################
    # Relative node properties #
    ############################

    def parent_nodes(self, vinds):
        """Get a list of parent nodes for specified vertices

        Parameters
        ----------
        vinds : Collection of ints
            Collection of vertex indices

        Returns
        -------
        numpy.array
            The parent node of each vertex index in vinds.
        """
        vinds_b = self.map_indices_to_unmasked(vinds)
        if isinstance(vinds_b, int):
            vinds_b = np.array([vinds_b])
        p_inds_filtered = self.filter_unmasked_indices_padded(
            self._rooted.parent_nodes(vinds_b)
        )
        return self.SkeletonIndex(p_inds_filtered)

    def cut_graph(self, vinds, directed=True, euclidean_weight=True):
        """Return a csgraph for the skeleton with specified vertices cut off from their parent vertex.

        Parameters
        ----------
        vinds :
            Collection of indices to cut off from their parent.
        directed : bool, optional
            Return the graph as directed, by default True
        euclidean_weight : bool, optional
            Return the graph with euclidean weights, by default True. If false, the unweighted.

        Returns
        -------
        scipy.sparse.csr.csr_matrix
            Graph with vertices in vinds cutt off from parents.
        """
        e_keep = ~np.isin(self.edges[:, 0], vinds)
        es_new = self.edges[e_keep]
        return utils.create_csgraph(
            self.vertices, es_new, euclidean_weight=euclidean_weight, directed=directed
        )

    def downstream_nodes(self, vinds, inclusive=True):
        """Get a list of all nodes downstream of a collection of indices

        Parameters
        ----------
        vinds : Collection of ints
            Collection of vertex indices

        Returns
        -------
        List of arrays
            List whose ith element is an array of all vertices downstream of the ith element of vinds.
        """
        vinds, return_single = utils.array_if_scalar(vinds)

        dns = []
        for vind in vinds:
            g = self.cut_graph(vind)
            d = sparse.csgraph.dijkstra(g.T, indices=[vind])
            if inclusive is False:
                d[vind] = np.inf
            dns.append(self.SkeletonIndex(np.flatnonzero(~np.isinf(d[0]))))

        if return_single:
            dns = dns[0]
        return dns

    def child_nodes(self, vinds):
        """Get a list of all immediate children of list of indices

        Parameters
        ----------
        vinds : Collection of ints
            Collection of vertex indices

        Returns
        -------
        List of arrays
            A list whose ith element is the children of the ith element of vinds.
        """
        vinds, return_single = utils.array_if_scalar(vinds)
        # return_single = False
        # if np.isscalar(vinds):
        #     vinds = [vinds]
        #     return_single = True
        # elif issubclass(type(vinds), np.ndarray):
        #     if len(vinds.shape) == 0:
        #         vinds = vinds.reshape(1)
        #         return_single = True

        cinds = []
        for vind in vinds:
            cinds.append(self.SkeletonIndex(self.edges[self.edges[:, 1] == vind, 0]))

        if return_single:
            cinds = cinds[0]
        return cinds

    #####################
    # Cover Path Functions #
    #####################

    def _compute_cover_paths(self, end_points=None):
        """Compute the list of cover paths along the skeleton"""
        cover_paths = []
        seen = np.full(self.n_vertices, False)
        if end_points is None:
            end_points = self.end_points

        ep_order = np.argsort(self.distance_to_root[end_points])[::-1]
        for ep in end_points[ep_order]:
            ptr = np.array(self.path_to_root(ep))
            cover_paths.append(self.SkeletonIndex(ptr[~seen[ptr]]))
            seen[ptr] = True
        return cover_paths

    @property
    def cover_paths(self):
        """list : List of numpy.array objects with self.n_end_points elements, each a rootward
        path (ordered set of indices) starting from an endpoint and continuing until it reaches
        a point on a path earlier on the list. Paths are ordered by end point distance from root,
        starting with the most distal. When traversed from begining to end, gives the longest rootward
        path down each major branch first. When traversed from end to begining, guarentees every
        branch point is visited at the end of all its downstream paths before being traversed along
        a path.
        """
        if self._cover_paths is None:
            self._cover_paths = self._compute_cover_paths()
        return self._cover_paths

    def cover_paths_specific(self, end_points):
        """Compute nonoverlapping paths from specified endpoints

        Parameters
        ----------
            end_points : array-like
                Array of skeleton vertices to use as end points

        Returns
        -------
            array
                List of cover paths using the specified end points. Note that this is not sorted in the same order
                (or necessarily the same length) as specified end points.
        """
        paths = self._compute_cover_paths(end_points=end_points)
        return [p for p in paths if len(p) > 0]

    ####################
    # Export functions #
    ####################

    def write_to_h5(self, filename, overwrite=True):
        """Write the skeleton to an HDF5 file. Note that this is done in the original dimensons, not the scaled dimensions.

        Parameters
        ----------
        filename : str
            Filename to save file
        overwrite : bool, optional
            Flag to specify whether to overwrite existing files, by default True
        """
        existing_voxel_scaling = self.voxel_scaling
        self.voxel_scaling = None
        skeleton_io.write_skeleton_h5(self, filename, overwrite=overwrite)
        self.voxel_scaling = existing_voxel_scaling

    def export_to_swc(
        self,
        filename,
        node_labels=None,
        radius=None,
        header=None,
        xyz_scaling=1000,
        resample_spacing=None,
        interp_kind="linear",
        tip_length_ratio=0.5,
    ):
        """
        Export a skeleton file to an swc file
        (see http://research.mssm.edu/cnic/swc.html for swc definition)

        Parameters
        ----------
        filename : str
            Name of the file to save the swc to
        node_labels : np.array
            None (default) or an interable of ints co-indexed with vertices.
            Corresponds to the swc node categories. Defaults to setting all
            odes to label 3, dendrite.
        radius : iterable
            None (default) or an iterable of floats. This should be co-indexed with vertices.
            Radius values are assumed to be in the same units as the node vertices.
        header : dict
            default None. Each key value pair in the dict becomes
            a parameter line in the swc header.
        xyz_scaling : Number
            default 1000. Down-scales spatial units from the skeleton's units to
            whatever is desired by the swc. E.g. nm to microns has scaling=1000.

        """
        skeleton_io.export_to_swc(
            self,
            filename,
            node_labels=node_labels,
            radius=radius,
            header=header,
            xyz_scaling=xyz_scaling,
            resample_spacing=resample_spacing,
            interp_kind=interp_kind,
            tip_length_ratio=tip_length_ratio,
        )


def resample(sk, spacing, kind="linear", tip_length_ratio=0.25):
    """Resample a skeleton's vertices

    Parameters
    ----------
    sk : Skeleton
        Input skeleton file with a skeleton
    spacing : numeric
        Desired spacing in nanometers
    kind : str, optional
        Type of interpolation to use when resampling. Options follow scipy.interpolate.interp1d. By default "linear"
    tip_length_ratio : float, optional
        The ratio of spacing to branch tip length that a branch tip must have in order to be included in the final skeleton
        for example: spacing is 10 and branch length is 8. do you want to include that final 8 length tip?
        then perhaps consider a tip_length_ratio of .75, by default 0.25

    Returns
    -------
    Skeleton
        New skeleton with resampled vertices.

    resample_map
        Array where the ith index corresponds to the ith vertex of the resampled skeleton and the value
        is the associated index in the original skeleton. To assign vertices, we assign a "domain" to each
        vertex in the original skeleton that is halfway between the vertex and its neighbors. Resampled
        vertices that fall within that domain (based on topology and distance-to-root) are then associated
        with the original vertex.
    """
    path_counter = 0
    branch_d = {}
    vert_list = []
    edge_list = []
    output_map_list = []

    for path in sk.cover_paths:
        new_verts, new_edges, output_map_path, branch_d = resample_path(
            path,
            sk,
            path_counter,
            spacing,
            kind,
            tip_length_ratio,
            branch_d,
        )
        vert_list.append(new_verts)
        edge_list.append(new_edges)
        output_map_list.append(output_map_path)
        path_counter += len(new_verts)

    new_verts = np.vstack(vert_list)
    new_edges = np.vstack(edge_list)
    resample_map = np.concatenate(output_map_list)

    return (
        Skeleton(
            new_verts,
            new_edges,
            root=branch_d[int(sk.root)],
            remove_zero_length_edges=False,
        ),
        resample_map,
    )


import numpy as np
from functools import partial
from scipy import interpolate
from scipy.spatial import KDTree


def make_windows(input_d):
    delta_d = np.abs(np.diff(input_d))
    d_min = input_d - np.concatenate([delta_d / 2, [0]])
    d_max = input_d + np.concatenate([[0], delta_d / 2])
    return d_min, d_max


def _assign_window(x, d_min, d_max):
    indv = np.flatnonzero(np.logical_and(x >= d_min, x < d_max))
    if len(indv) == 1:
        return indv[0]
    else:
        return None


def assign_windows(des_d, init_d):
    "Assign desired distances to windows determined from the initial distances for each vertex"
    d_min, d_max = make_windows(init_d)
    _p_assign_window = partial(_assign_window, d_min=d_min, d_max=d_max)
    _assign_window_v = np.vectorize(_p_assign_window)
    inds = _assign_window_v(des_d)
    return inds


def resample_path(path, sk, path_counter, spacing, kind, tip_length_ratio, branch_d):
    "Resample a specific path to get a desired spacing"
    if path[-1] != sk.root:
        # find that last edge whose start point was the first vertex in the path
        # do this so we don't get big gaps from soma->branch or branch->branch
        last_node = int(sk.parent_nodes(path[-1]))
        path_as_list = list(path)
        path_as_list.append(last_node)
        mod_path = sk.SkeletonIndex(path_as_list)
        add_last_edge = True
    else:
        mod_path = path
        add_last_edge = False

    # use the distance from root to parameterize the path
    input_d = sk.distance_to_root[mod_path]
    # the desired distances from root are evenly spaced according to spacing
    des_d = np.arange(np.min(input_d), np.max(input_d), spacing)

    # setup an interpolation function based upon distance to root as input and xyz as output
    fi = interpolate.interp1d(input_d, sk.vertices[mod_path, :], kind=kind, axis=0)
    # use the function to interpolate the new values
    # we need to add that last node to the verts ONLY IF the last edge length
    # meets the cutoff defined in tip_length_ratio
    new_verts = fi(des_d)

    inds = assign_windows(des_d, input_d)
    output_map_path = np.array(mod_path[inds])

    tip_ind = mod_path[0]
    true_tip = sk.vertices[tip_ind, :]
    if np.linalg.norm(new_verts[-1] - true_tip) / spacing > tip_length_ratio:
        new_verts = np.vstack([new_verts, true_tip])
        output_map_path = np.concatenate((output_map_path, [tip_ind]))

    # find the index of the old branch points in the new path
    is_branch = np.isin(np.array(path), sk.branch_points)
    path_branch = path[is_branch]
    path_branch_verts = sk.vertices[path_branch, :]
    tree = KDTree(new_verts)
    map_ds, new_branch_on_path = tree.query(path_branch_verts)
    new_branch_on_path += path_counter
    # create a temporary dictionary with this path's branch  points
    new_branch_d = {pb: nw for pb, nw in zip(path_branch, new_branch_on_path)}
    # update the overall mapping dictionary
    branch_d.update(new_branch_d)

    # new edges just march down path from last vertex to first
    new_edges = (
        np.vstack(
            [
                np.arange(len(new_verts) - 1, 0, -1),
                np.arange(len(new_verts) - 2, -1, -1),
            ]
        ).T
        + path_counter
    )
    if add_last_edge:
        # if we do have one, then we want to add an edge that is the from the first vertex
        # in the path, to the new vertex (mapped through branch_d) of the edge we found
        # in the original edge list
        new_edges = np.vstack([new_edges, [path_counter, branch_d[last_node]]])
    return new_verts, new_edges, output_map_path, branch_d


from .utils import decompress_mesh_data
from ..skeleton import Skeleton
from ..trimesh_io import Mesh
import h5py
import os
import orjson
import pandas as pd
import warnings
import numpy as np
from dataclasses import asdict

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

NULL_VERSION = 1
LATEST_VERSION = 2


def save_meshwork_metadata(filename, mw, version=LATEST_VERSION):
    with h5py.File(filename, "a") as f:
        f.attrs["voxel_resolution"] = mw.anno.voxel_resolution
        f.attrs["version"] = version
        if mw.seg_id is not None:
            f.attrs["seg_id"] = mw.seg_id


def load_meshwork_metadata(filename):
    meta = {}
    with h5py.File(filename, "r") as f:
        meta["seg_id"] = f.attrs.get("seg_id", None)
        meta["voxel_resolution"] = f.attrs.get("voxel_resolution", None)
        meta["version"] = f.attrs.get("version", NULL_VERSION)
    return meta


def save_meshwork_mesh(filename, mw, version=LATEST_VERSION):
    node_mask = mw.mesh_mask
    if mw._original_mesh_data is not None:
        vs, fs, es, nm, vxsc = decompress_mesh_data(*mw._original_mesh_data)
        mesh = Mesh(vs, fs, link_edges=es, node_mask=nm, voxel_scaling=vxsc)
    else:
        mesh = mw.mesh

    with h5py.File(filename, "a") as f:
        f.create_group("mesh")
        f.create_dataset("mesh/vertices", data=mesh.vertices, compression="gzip")
        f.create_dataset("mesh/faces", data=mesh.faces, compression="gzip")
        f.create_dataset("mesh/node_mask", data=mesh.node_mask, compression="gzip")
        if mesh.voxel_scaling is not None:
            f["mesh"].attrs["voxel_scaling"] = mesh.voxel_scaling
        if mesh.link_edges is not None:
            f.create_dataset(
                "mesh/link_edges", data=mesh.link_edges, compression="gzip"
            )
        f.create_dataset("mesh/mesh_mask", data=node_mask, compression="gzip")


def load_meshwork_mesh(filename, version=NULL_VERSION):
    with h5py.File(filename, "r") as f:
        verts = f["mesh/vertices"][()]
        faces = f["mesh/faces"][()]
        if len(faces.shape) == 1:
            faces = faces.reshape(-1, 3)

        if "link_edges" in f["mesh"].keys():
            link_edges = f["mesh/link_edges"][()]
        else:
            link_edges = None

        node_mask = f["mesh/node_mask"][()]
        voxel_scaling = f["mesh"].attrs.get("voxel_scaling", None)
        mesh_mask = f["mesh/mesh_mask"][()]
    return (
        Mesh(
            vertices=verts,
            faces=faces,
            link_edges=link_edges,
            node_mask=node_mask,
            voxel_scaling=voxel_scaling,
        ),
        mesh_mask,
    )


def save_meshwork_skeleton(filename, mw, version=LATEST_VERSION):
    if mw.skeleton is None:
        return

    sk = mw.skeleton.reset_mask()
    with h5py.File(filename, "a") as f:
        f.create_group("skeleton")
        f.create_dataset("skeleton/vertices", data=sk.vertices, compression="gzip")
        f.create_dataset("skeleton/edges", data=sk.edges, compression="gzip")
        f.create_dataset("skeleton/root", data=sk.root)
        f.create_dataset(
            "skeleton/meta",
            data=np.string_(
                orjson.dumps(asdict(sk.meta), option=orjson.OPT_SERIALIZE_NUMPY)
            ),
        )

        f.create_dataset(
            "skeleton/mesh_to_skel_map", data=sk.mesh_to_skel_map, compression="gzip"
        )
        if sk.radius is not None:
            f.create_dataset("skeleton/radius", data=sk.radius, compression="gzip")
        if sk.mesh_index is not None:
            f.create_dataset(
                "skeleton/mesh_index", data=sk.mesh_index, compression="gzip"
            )
        if sk.voxel_scaling is not None:
            f["skeleton"].attrs["voxel_scaling"] = sk.voxel_scaling


def load_meshwork_skeleton(filename, version=NULL_VERSION):
    with h5py.File(filename, "r") as f:
        if "skeleton" not in f:
            return None

        verts = f["skeleton/vertices"][()]
        edges = f["skeleton/edges"][()]
        root = f["skeleton/root"][()]
        mesh_to_skel_map = f["skeleton/mesh_to_skel_map"][()]
        if "radius" in f["skeleton"].keys():
            radius = f["skeleton/radius"][()]
        else:
            radius = None
        voxel_scaling = f["skeleton"].attrs.get("voxel_scaling", None)

        if "meta" in f["skeleton"].keys():
            meta = orjson.loads(f["skeleton/meta"][()].tobytes())
        else:
            meta = {}

        if "mesh_index" in f["skeleton"].keys():
            mesh_index = f["skeleton/mesh_index"][()]
        else:
            mesh_index = None
        return Skeleton(
            verts,
            edges,
            root=root,
            radius=radius,
            mesh_to_skel_map=mesh_to_skel_map,
            mesh_index=mesh_index,
            voxel_scaling=voxel_scaling,
            meta=meta,
        )


def _save_dataframe_generic(df, table_name, filename):
    key = f"annotations/{table_name}/data"
    dat = orjson.dumps(
        df.to_dict(), option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY
    )
    with h5py.File(filename, "a") as f:
        f.create_dataset(key, data=np.string_(dat))
    pass


def _load_dataframe_generic(filename, table_name):
    key = f"annotations/{table_name}/data"
    with h5py.File(filename, "r") as f:
        dat = f[key][()].tobytes()
        df = pd.DataFrame.from_records(orjson.loads(dat))
    return df


def _save_dataframe_pandas(data, table_name, filename):
    data.to_hdf(
        filename, f"annotations/{table_name}/data", complib="blosc", complevel=5
    )
    pass


def _load_dataframe_pandas(filename, table_name):
    return pd.read_hdf(filename, f"annotations/{table_name}/data")


anno_save_function = {1: _save_dataframe_pandas, 2: _save_dataframe_generic}
anno_load_function = {1: _load_dataframe_pandas, 2: _load_dataframe_generic}


def save_meshwork_annotations(filename, mw, version=LATEST_VERSION):
    annos = mw.anno
    if version not in anno_save_function:
        raise ValueError(f"Version must be one of {list(anno_save_function)}")

    for table_name in annos.table_names:
        with h5py.File(filename, "a") as f:
            dset = f.create_group(f"annotations/{table_name}")
            anno = annos[table_name]
            dset.attrs["anchor_to_mesh"] = int(anno.anchored)
            if anno.point_column is not None:
                dset.attrs["point_column"] = anno.point_column
            dset.attrs["max_distance"] = anno._max_distance
            dset.attrs["defined_index"] = int(anno._defined_index)
            if anno._defined_index is True:
                dset.attrs["index_column"] = anno.index_column_original
            
        anno_save_function[version](
            annos[table_name].data_original, table_name, filename
        )


def load_meshwork_annotations(filename, version=NULL_VERSION):

    with h5py.File(filename, "r") as f:
        if "annotations" not in f:
            return {}
        table_names = list(f["annotations"].keys())

    annotation_dfs = {}
    for table_name in table_names:
        annotation_dfs[table_name] = {}
        annotation_dfs[table_name]["data"] = anno_load_function[version](
            filename, table_name
        )
        with h5py.File(filename, "r") as f:
            dset = f[f"annotations/{table_name}"]
            annotation_dfs[table_name]["anchor_to_mesh"] = bool(
                dset.attrs.get("anchor_to_mesh")
            )
            annotation_dfs[table_name]["point_column"] = dset.attrs.get(
                "point_column", None
            )
            annotation_dfs[table_name]["max_distance"] = dset.attrs.get("max_distance")
            if bool(dset.attrs.get("defined_index", False)):
                annotation_dfs[table_name]["index_column"] = dset.attrs.get(
                    "index_column", None
                )
    return annotation_dfs


def _save_meshwork(filename, mw, overwrite=False, version=LATEST_VERSION):
    if isinstance(filename, str):
        if os.path.exists(filename):
            if overwrite is False:
                raise FileExistsError()
            else:
                print(f"\tDeleting existing data in {filename}...")
                with h5py.File(filename, "w") as f:
                    pass

    save_meshwork_metadata(filename, mw, version=version)
    save_meshwork_mesh(filename, mw, version=version)
    save_meshwork_skeleton(filename, mw, version=version)
    save_meshwork_annotations(filename, mw, version=version)


def _load_meshwork(filename):
    meta = load_meshwork_metadata(filename)
    version = meta.get("version")
    mesh, mask = load_meshwork_mesh(filename, version=version)
    skel = load_meshwork_skeleton(filename, version=version)
    annos = load_meshwork_annotations(filename, version=version)
    return meta, mesh, skel, annos, mask



try:
    from .. import trimesh_vtk

    _vtk_loaded = True
except:
    _vtk_loaded = False

from ..skeleton_io import swc_node_labels
from ..trimesh_io import Mesh
from ..skeleton import Skeleton, resample

import contextlib
import io
import pandas as pd
import numpy as np
from scipy import sparse
from functools import wraps
from .utils import (
    DEFAULT_VOXEL_RESOLUTION,
    MeshworkIndexFactory,
    in1d_items,
    in1d_first_item,
    unique_column_name,
    compress_mesh_data,
    decompress_mesh_data,
    MaskedMeshMemory,
    window_matrix,
)
from . import meshwork_io


class AnchoredAnnotationManager(object):
    def __init__(
        self,
        anchor_mesh=None,
        filter_mesh=None,
        voxel_resolution=None,
    ):
        """Collection of dataframes anchored to a common mesh and filter.

        Parameters
        ----------
        mesh : trimesh_io.Mesh, optional
            Mesh to use to link points to vertex indices, by default None
        voxel_resolution : array-like, optional
            3-element array of the resolution of voxels in the point columns of the dataframes, by default None
        """
        if voxel_resolution is None:
            voxel_resolution = DEFAULT_VOXEL_RESOLUTION
        self._voxel_resolution = np.array(voxel_resolution).reshape((1, 3))
        if anchor_mesh is None:
            self._anchor_mesh = None
        else:
            self._anchor_mesh = MaskedMeshMemory(anchor_mesh)
        self._filter_mesh = filter_mesh
        self._MeshIndex = np.array
        self._data_tables = dict()

    def __len__(self):
        return len(self._data_tables)

    def __getitem__(self, key):
        return self._data_tables[key]

    def __delitem__(self, key):
        if key in self._data_tables:
            del self._data_tables[key]
            del self.__dict__[key]

    def __repr__(self):
        return f"Data tables: {self.table_names}"

    def items(self):
        return self._data_tables.items()

    @property
    def table_names(self):
        """List of data table names"""
        return list(self._data_tables.keys())

    @property
    def voxel_resolution(self):
        """Resolution in nm of the data in the point column"""
        return self._voxel_resolution

    @voxel_resolution.setter
    def voxel_resolution(self, new_res):
        self._voxel_resolution = np.array(new_res).reshape((1, 3))
        for tn in self.table_names:
            self._data_tables[tn].voxel_resolution = self._voxel_resolution

    def update_anchor_mesh(self, new_mesh):
        """Change or add the mesh to use for proximity-linking"""

        self._anchor_mesh = new_mesh

        for name in self.table_names:
            table = self._data_tables[name]
            if table.anchored is False:
                continue

            if table._index_column_base in table._original_columns:
                index_column = table._index_column_base
            else:
                index_column = None

            if table._point_column in table._original_columns:
                point_column = table._point_column
            else:
                point_column = None

            self._data_tables[name] = AnchoredAnnotation(
                name,
                table.data_original,
                self._anchor_mesh,
                point_column=point_column,
                max_distance=table._max_distance,
                index_column=index_column,
                voxel_resolution=self.voxel_resolution,
            )
            if self._filter_mesh is not None:
                self._data_tables[name]._filter_data(self._filter_mesh)

    def _add_attribute(self, key):
        if key in dir(self):
            if not isinstance(self.__getattribute__(key), AnchoredAnnotation):
                return
        else:
            self.__dict__[key] = self._data_tables[key]

    def add_annotations(
        self,
        name,
        data,
        mask=False,
        point_array=False,
        anchored=True,
        point_column=None,
        max_distance=np.inf,
        index_column=None,
        overwrite=False,
        voxel_resolution=None,
    ):
        "Add a dataframe to the manager"

        if name in self.table_names and overwrite is False:
            raise ValueError(
                "Table name already taken. Overwrite or choose a different name."
            )
        else:
            self.remove_annotations(name)

        if voxel_resolution is None:
            voxel_resolution = self.voxel_resolution

        if issubclass(type(data), np.ndarray):
            if len(data.shape) == 2:
                if data.shape[1] == 3:
                    data = pd.DataFrame(data={"pt": data.tolist()})
                    point_column = "pt"
                else:
                    raise ValueError("Arrays must be of size Nx3")

        self._data_tables[name] = AnchoredAnnotation(
            name,
            data,
            self._anchor_mesh,
            point_column=point_column,
            mask=mask,
            point_array=point_array,
            anchor_to_mesh=anchored,
            max_distance=max_distance,
            index_column=index_column,
            voxel_resolution=voxel_resolution,
        )
        self._data_tables[name]._register_MeshIndex(self._MeshIndex)
        self._add_attribute(name)

    def remove_annotations(self, name):
        "Remove a data table from the manager"
        if isinstance(name, str):
            name = [name]
        for n in name:
            del self[n]

    def anchor_annotations(self, name):
        "If an annotation is not anchored, link it to the current anchor mesh and apply the current filters"
        if isinstance(name, str):
            name = [name]
        for n in name:
            self._data_tables[n]._anchor_to_mesh(self._anchor_mesh)
            if self._filter_mesh is not None:
                self._data_tables[n]._filter_data(self._filter_mesh)

    def filter_annotations(self, new_mesh):
        "Use a masked mesh to filter all anchored annotations"
        self._filter_mesh = MaskedMeshMemory(new_mesh)
        for tn in self.table_names:
            self._data_tables[tn]._filter_data(self._filter_mesh)

    def remove_filter(self):
        "Remove filters from the annotations"
        self._filter_mesh = None
        for tn in self.table_names:
            self._data_tables[tn]._reset_filter()

    def _register_MeshIndex(self, NewMeshIndex):
        self._MeshIndex = NewMeshIndex
        for tn in self.table_names:
            self._data_tables[tn]._register_MeshIndex(NewMeshIndex)


class AnchoredAnnotation(object):
    def __init__(
        self,
        name,
        data,
        mesh=None,
        mask=False,
        point_array=False,
        anchor_to_mesh=True,
        point_column=None,
        max_distance=np.inf,
        index_column=None,
        voxel_resolution=None,
        voxel_scaling=None,
    ):
        if point_array is True:
            if isinstance(data, np.ndarray):
                data = data.tolist()
            if point_column is None:
                point_column = "position"
            data = pd.DataFrame({point_column: data})
        elif mask == True:
            if isinstance(data, np.ndarray):
                data = data.ravel()
            index_column = "mesh_index"
            data = pd.DataFrame({index_column: data})

        self._name = name
        self._data = data.reset_index(drop=True)
        self._original_columns = data.columns
        self._max_distance = max_distance

        self._point_column = point_column
        if index_column is None:
            defined_index = False
            index_column = unique_column_name(None, "mesh_index", data)
        else:
            defined_index = True
        self._defined_index = defined_index
        self._index_column_base = index_column
        self._index_column_filt = f"{index_column}_filt"

        # Initalize to -1 so the column exists
        self._data[self._index_column_base] = -1
        self._data[self._index_column_filt] = -1

        if self._defined_index:
            if mesh is not None and anchor_to_mesh is True:
                self._data[self._index_column_base] = mesh.map_indices_to_unmasked(
                    data[index_column]
                )
                self._data[self._index_column_filt] = data[index_column].values

        if self._defined_index:
            self._orig_col_plus_index = list(self._original_columns)
        else:
            self._orig_col_plus_index = list(self._original_columns) + [
                self._index_column_filt
            ]

        valid_column = unique_column_name(index_column, "valid", data)
        self._data[valid_column] = True
        self._valid_column = valid_column

        self._mask_column = unique_column_name(index_column, "in_mask", data)
        # Initalize in_mask to True before any subsequent masking
        self._data[self._mask_column] = True

        if voxel_resolution is None:
            voxel_resolution = DEFAULT_VOXEL_RESOLUTION
        if mesh.voxel_scaling is not None:
            voxel_resolution = voxel_resolution * mesh.voxel_scaling

        self._voxel_resolution = np.array(voxel_resolution).reshape((1, 3))
        self._MeshIndex = None

        self._anchor_mesh = None
        self._filter_mesh = None
        self._anchored = anchor_to_mesh
        if self._anchored and mesh is not None:
            self._anchor_points(mesh)

    def __repr__(self):
        return self.df.__repr__()

    def __getitem__(self, key):
        return self.df.__getitem__(key)

    def __len__(self):
        return len(self.df)

    @property
    def name(self):
        return self._name

    @property
    def point_column(self):
        return self._point_column

    @property
    def index_column(self):
        return self._index_column_filt

    @property
    def index_column_original(self):
        return self._index_column_base

    @property
    def _is_valid(self):
        return self._data[self._valid_column]

    @property
    def _in_mask(self):
        return self._data[self._mask_column]

    @property
    def _is_included(self):
        return np.logical_and(self._is_valid, self._in_mask)

    @property
    def df(self):
        if self.anchored:
            return self._data[self._orig_col_plus_index][self._is_included]
        else:
            return self._data[self._original_columns][self._is_included]

    @property
    def voxel_resolution(self):
        return self._voxel_resolution

    @voxel_resolution.setter
    def voxel_resolution(self, new_res):
        self._voxel_resolution = np.array(new_res).reshape((1, 3))

    @property
    def voxels(self):
        if self.point_column is None or len(self.df) == 0:
            return np.zeros((0, 3))
        else:
            return np.vstack(self.df[self.point_column].values)

    @property
    def points(self):
        if self.point_column is None:
            return np.zeros((0, 3))
        else:
            return self.voxels * self.voxel_resolution

    def _register_MeshIndex(self, NewClass):
        self._MeshIndex = NewClass

    @property
    def MeshIndex(self):
        if self._MeshIndex is None:
            return np.array
        return self._MeshIndex

    @property
    def mesh_index(self):
        if self.anchored:
            return self.MeshIndex(
                self._data[self._index_column_filt][self._is_included].values
            )
        else:
            return None

    @property
    def skel_index(self):
        return self.mesh_index.to_skel_index

    @property
    def mesh_mask(self):
        return self.mesh_index.to_mesh_mask

    @property
    def skel_mask(self):
        return self.mesh_index.to_skel_mask

    @property
    def _mesh_index_base(self):
        return self._data[self._index_column_base].values

    @property
    def data_original(self):
        return self._data[self._original_columns]

    def _anchor_points(self, mesh):
        if self._defined_index is False:
            dist, minds_filt = mesh.kdtree.query(self.points)
            self._data[self._index_column_filt] = minds_filt

            minds_base = mesh.map_indices_to_unmasked(minds_filt)
            self._data[self._index_column_base] = minds_base
            self._data[self._valid_column] = dist < self._max_distance
        else:
            self._data[self._valid_column] = True

        self._anchor_mesh = MaskedMeshMemory(mesh, index_only=True)
        self._filter_mesh = self._anchor_mesh

    def _filter_data(self, filter_mesh):
        """Get the subset of data points that are associated with the mesh"""
        if self._anchored:
            self._data[self._mask_column] = filter_mesh.node_mask[self._mesh_index_base]
            self._data[
                self._index_column_filt
            ] = filter_mesh.filter_unmasked_indices_padded(self._mesh_index_base)
            self._filter_mesh = MaskedMeshMemory(filter_mesh)

    def _filter_query_response(self, row_filter):
        _parent = self

        class _FilterQueryResponse(object):
            def __init__(self, row_filter):
                self._row_filter = row_filter

            @property
            def row_filter(self):
                return self._row_filter

            @property
            def voxels(self):
                return _parent.voxels[self.row_filter]

            @property
            def points(self):
                return _parent.points[self.row_filter]

            @property
            def df(self):
                return _parent.df[self.row_filter]

            @property
            def count(self):
                return np.sum(row_filter)

            @property
            def mesh_index(self):
                return _parent.MeshIndex(
                    _parent._data[_parent._index_column_filt][self.row_filter].values
                )

        return _FilterQueryResponse(row_filter)

    def _filter_query(self, node_mask):
        """Returns the data contained with a given filter without changing any indexing."""
        node_mask_base = self._filter_mesh.map_boolean_to_unmasked(node_mask)
        if self._anchored:
            keep_rows = node_mask_base[self._mesh_index_base]
            return keep_rows[self._is_included]
        else:
            return np.full(len(self.df), True)

    def filter_query(self, node_mask):
        row_filter = self._filter_query(node_mask)
        return self._filter_query_response(row_filter)

    def _reset_filter(self):
        if self._anchored:
            self._data[self._mask_column] = True
            self._data[
                self._index_column_filt
            ] = self._anchor_mesh.filter_unmasked_indices(self._mesh_index_base)
            self._filter_mesh = self._anchor_mesh

    def _anchor_to_mesh(self, anchor_mesh):
        self._anchored = True
        self._reset_filter()
        self._data[self._valid_column] = True
        self._anchor_points(anchor_mesh)

    @property
    def anchored(self):
        return self._anchored


class Meshwork(object):
    """Class to integrate mesh, skeleton, and annotation data

    Parameters
    ----------
    mesh : trimesh_io.Mesh
        Mesh in a form for skeletonization and attaching synapses
    skeleton : skeleton.Skeleton, optional
        Skeleton built from the same mesh as Mesh.
    seg_id : int or None, optional
        Segment id for object
    voxel_resolution : array-like or None, optional
        3-element voxel resolution (if None, [4,4,40]) to convert between
        annotation point units (voxels) and mesh vertex units (e.g. nanometers).
    """

    def __init__(
        self,
        mesh=None,
        skeleton=None,
        seg_id=None,
        voxel_resolution=None,
        voxel_scaling=None,
    ):
        if mesh is None and skeleton is None:
            raise ValueError("Must include at least one of mesh or skeleton")

        if mesh is None and skeleton is not None:
            mesh = Mesh(
                vertices=skeleton.vertices,
                faces=np.zeros((0, 3)),
                link_edges=skeleton.edges,
            )

        self._seg_id = seg_id
        self._mesh = mesh
        self._skeleton = skeleton

        if voxel_resolution is None:
            voxel_resolution = DEFAULT_VOXEL_RESOLUTION
        self._anno = AnchoredAnnotationManager(
            self._mesh, voxel_resolution=voxel_resolution
        )

        self._voxel_scaling = voxel_scaling
        self._mesh.voxel_scaling = voxel_scaling
        if skeleton is not None:
            self._skeleton.voxel_scaling = voxel_scaling

        self._original_mesh_data = None
        self._MeshIndex = None
        self._SkeletonIndex = None
        self._recompute_indices()

    @property
    def seg_id(self):
        """Segmentation id for the object"""
        return self._seg_id

    @property
    def voxel_scaling(self):
        return self._voxel_scaling

    @voxel_scaling.setter
    def voxel_scaling(self, new_scaling):
        self._mesh.voxel_scaling = new_scaling
        if self.skeleton is not None:
            self._skeleton.voxel_scaling = new_scaling
        self._recompute_indices()

    def __copy__(self):
        with io.BytesIO() as bio:
            self.save_meshwork(bio)
            bio.seek(0)
            nrn_copy = load_meshwork(bio)
        return nrn_copy

    def copy(self):
        return self.__copy__()

    ##################
    # Mesh functions #
    ##################

    def _recompute_indices(self):
        self._MeshIndex, self._SkeletonIndex = MeshworkIndexFactory(self)
        self.anno._register_MeshIndex(self._MeshIndex)
        if self.skeleton is not None:
            self.skeleton._register_skeleton_index(self._SkeletonIndex)

    @property
    def MeshIndex(self):
        """Numpy array-like object for mesh indices with handy conversion features"""
        if self._MeshIndex is None:
            self._recompute_indices()
        return self._MeshIndex

    def _convert_to_meshindex(self, mesh_indices):
        if isinstance(mesh_indices, self.MeshIndex):
            return mesh_indices
        elif isinstance(mesh_indices, self.SkeletonIndex):
            return mesh_indices.to_mesh_index
        else:
            if np.isscalar(mesh_indices):
                mesh_indices = np.array([mesh_indices], dtype=int)
            return self.MeshIndex(mesh_indices)

    @property
    def SkeletonIndex(self):
        """Numpy array-like object for skeleton indices with handy conversion features"""
        if self._SkeletonIndex is None:
            self._recompute_indices()
        return self._SkeletonIndex

    def _convert_to_skelindex(self, skel_indices):
        if isinstance(skel_indices, self.SkeletonIndex):
            return skel_indices
        elif isinstance(skel_indices, self.MeshIndex):
            return skel_indices.to_skel_index
        else:
            if np.isscalar(skel_indices):
                skel_indices = np.array([skel_indices], dtype=int)
            return self.SkeletonIndex(skel_indices)

    def _reset_indices(self):
        self._MeshIndex = None
        self._SkeletonIndex = None
        if self.skeleton is not None:
            self.skeleton._register_skeleton_index(self.SkeletonIndex)
        self.anno._register_MeshIndex(self.MeshIndex)

    @property
    def mesh(self):
        """Copy of the neuronal mesh"""
        return self._mesh

    @property
    def mesh_mask(self):
        """The mesh mask, a boolean array the length of the number of mesh vertices."""
        return self.mesh.node_mask

    @property
    def vertices(self):
        return self.mesh.vertices

    @property
    def edges(self):
        return self.mesh.edges

    @property
    def faces(self):
        return self.mesh.faces

    def apply_mask(self, mask):
        """Apply a mesh mask to the meshwork object

        Parameters
        ----------
        mask : array of booleans
            Array with the same number of elements as mesh vertices. True elements are kept,
            False elements are masked out.
        """
        if self._original_mesh_data is None:
            self._original_mesh_data = compress_mesh_data(self.mesh)

        if self.skeleton is not None:
            sk_mask = self._mesh_mask_to_skel_mask(mask)
            self.skeleton.apply_mask(sk_mask, in_place=True)

        self._mesh = self.mesh.apply_mask(mask)
        self._anno.filter_annotations(self.mesh)
        self._reset_indices()

    def reset_mask(self, to=None):
        """Remove mask and restore object to its original state.

        Parameters
        ----------
        to : array, optional
            Mask array of booleans to apply after resetting the mask. Equivalent to reset_mask followed by apply_mask.
        """
        if self._original_mesh_data is not None:
            self._anno.remove_filter()

            vs, fs, es, nm, vxsc = decompress_mesh_data(*self._original_mesh_data)
            self._mesh = Mesh(vs, fs, link_edges=es, node_mask=nm, voxel_scaling=vxsc)

            self._original_mesh_data = None
            if self.skeleton is not None:
                self._skeleton.reset_mask(in_place=True)
            self._reset_indices()

            if to is not None:
                self.apply_mask(to)

    @contextlib.contextmanager
    def mask_context(self, mask):
        """Use with-statement context to temporarily act on a masked state of an object.

        Parameters
        ----------
        mask : array or None,
            A boolean array with the same number of elements as mesh vertices. True elements are kept, False are masked out. If None, resets the mask entirely.
        """
        current_mask = self.mesh_mask
        try:
            if mask is None:
                self.reset_mask()
            else:
                self.apply_mask(mask)
            yield self
        finally:
            self.reset_mask(to=current_mask)

    ##################
    # Anno functions #
    ##################

    @property
    def anno(self):
        """AnchoredAnnotationManager associated with the anchor mesh"""
        return self._anno

    def add_annotations(
        self,
        name,
        data,
        anchored=True,
        mask=False,
        point_array=False,
        point_column=None,
        max_distance=np.inf,
        index_column=None,
        overwrite=False,
    ):
        """Associate annotation data with the mesh

        Parameters
        ----------
        name : str
            Name of the annotation data, to be used as a unique key.
        data : pandas.DataFrame
            DataFrame containing annotation data.
        anchored : bool, optional
            If True, associates locations in a specified point column with mesh vertices.
        mask : bool, optional
            If True, assumes the data is a boolean with the same length as the mesh vertices and converts True values to mesh indices. The resulting annotation does not keep track of points, only indices.
        point_array : bool, optional
            If True, assumes the data is a list or array of 3d points and constructs a basic dataframe to hold them (using the point_column value if provided).
        point_column : str or None, optional
            Column name holding 3-element point position in voxel units. Must be specified for anchored annotations.
        max_distance : numeric, optional
            Maximum distance between annotation point and mesh vertex to connect. If a row is
            farther than this distance, it is silently filtered out of the annotation data. By default, np.inf.
        index_column : str or None, optional
            If mesh indices are already computed, entries here are used instead of closest-distance.
        overwrite : bool, optional
            If True, overwrite an existing annotation with the same name.
        """
        self._anno.add_annotations(
            name=name,
            data=data,
            anchored=anchored,
            mask=mask,
            point_array=point_array,
            point_column=point_column,
            max_distance=max_distance,
            index_column=index_column,
            overwrite=overwrite,
        )

    def remove_annotations(self, name):
        """Remove annotation from AnnotationManager

        Parameters
        ----------
        name : str
            Name of the annotation to remove
        """
        self._anno.remove_annotations(name)

    def anchor_annotations(self, name):
        """Set an unanchored annotation to an anchored one

        Parameters
        ----------
        name : str
            Name of the annotation data to anchor
        """
        self._anno.anchor_annotations(name)

    ######################
    # Skeleton functions #
    ######################

    class OnlyIfSkeleton(object):
        """Convenience class to handle situations without skeletons"""

        @staticmethod
        def exists(func):
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                if self.skeleton is not None:
                    return func(self, *args, **kwargs)
                else:
                    return None

            return wrapper

    @property
    def skeleton(self):
        """Associated skeleton"""
        return self._skeleton

    def skeletonize_mesh(
        self,
        soma_pt=None,
        collapse_soma=True,
        soma_thresh_distance=7500,
        invalidation_distance=12000,
        compute_radius=True,
        shape_function="single",
        overwrite=False,
        collapse_function="sphere",
        collapse_params={},
    ):
        """Skeletonize the anchor mesh. Always uses the mesh as used to initialize the class instance.

        Parameters
        ----------
        soma_pt : arrary or None, optional
            3-element point (in mesh coordinates) specifying the center of the soma.
        soma_thresh_distance : numeric, optional
            Distance threshold to consider all inside the soma. Default is 7500.
        invalidation_distance : int, optional
            Invalidation distance for the TEASAR skeletonization. Default is 12000.
        compute_radius : bool, optional
            If True, computes the "shape distance function" giving the diameter of the mesh at the skeleton point.
        shape_function: 'single' or 'cone', optional
            Determines whether to use a single ray or a cone of rays to determine mesh diameter.
        overwrite : bool, optional
            If True, will run even if an existing skeletonization is present.
        """
        from meshparty.skeletonize import skeletonize_mesh

        if self._original_mesh_data is not None:
            vs, fs, es, nm, vxsc = decompress_mesh_data(*self._original_mesh_data)
            mesh_to_sk = Mesh(vs, fs, link_edges=es, node_mask=nm, voxel_scaling=vxsc)
        else:
            mesh_to_sk = self.mesh

        if self._skeleton is None or overwrite is True:
            self._skeleton = skeletonize_mesh(
                mesh_to_sk,
                soma_pt=soma_pt,
                soma_radius=soma_thresh_distance,
                collapse_soma=collapse_soma,
                invalidation_d=invalidation_distance,
                compute_original_index=True,
                compute_radius=compute_radius,
                shape_function=shape_function,
                collapse_function=collapse_function,
                collapse_params=collapse_params,
                meta={"root_id": self.seg_id},
            )
            self._reset_indices()
        else:
            print("Skeleton already exists")
        pass

    # all functions of this group take filtered indices and return filtered indices.
    def _mind_to_skind_padded(self, minds):
        minds_b = self.mesh.map_indices_to_unmasked(minds)
        skinds = self.skeleton.mesh_to_skel_map[minds_b]
        return skinds

    def _mind_to_skind(self, minds):
        mind_padded = self._mind_to_skind_padded(minds)
        return mind_padded[mind_padded >= 0]

    def _mesh_mask_to_skel_mask(self, mesh_mask):
        mesh_mask = self.mesh.map_boolean_to_unmasked(mesh_mask)
        skel_inds = np.unique(self.skeleton.mesh_to_skel_map_base[mesh_mask])
        skel_mask = np.full(self.skeleton.unmasked_size, False)
        skel_mask[skel_inds] = True
        return skel_mask

    def _skind_to_mind_mask_base(self, skinds):
        skinds_b = self.skeleton.map_indices_to_unmasked(skinds)
        minds_b_assoc = np.isin(self.skeleton.mesh_to_skel_map_base, skinds_b)
        return minds_b_assoc

    def _skind_to_mind_mask(self, skinds):
        return self.mesh.filter_unmasked_boolean(self._skind_to_mind_mask_base(skinds))

    def _skind_to_mind_index(self, skinds):
        return np.flatnonzero(self._skind_to_mind_mask(skinds))

    def _skind_regions(self, skinds):
        out = in1d_items(self.skeleton.mesh_to_skel_map[self.mesh.node_mask], skinds)
        return out

    def _skind_region_first(self, skinds):
        if isinstance(skinds, int):
            skinds = [skinds]
        if issubclass(type(skinds), np.ndarray):
            if len(skinds.shape) == 0:
                skinds = skinds.reshape(1)
        return in1d_first_item(
            self.skeleton.mesh_to_skel_map[self.mesh.node_mask], skinds
        )

    @OnlyIfSkeleton.exists
    def skeleton_property_to_mesh(
        self, skeleton_property, mesh_inds=None, no_map_value=-1.0
    ):
        """Map a property at skeleton points to the mesh, converting nodes
        without a clear mapping to a specified value (default -1).
        """
        if mesh_inds is None:
            mesh_inds = np.arange(self.mesh.n_vertices)
        mesh_inds = self._convert_to_meshindex(mesh_inds)
        mesh_property = np.full(len(mesh_inds), no_map_value)
        skids = mesh_inds.to_skel_index_padded
        mesh_property[skids >= 0] = skeleton_property[skids[skids >= 0]]
        return mesh_property

    @OnlyIfSkeleton.exists
    def mesh_property_to_skeleton(
        self,
        mesh_property,
        aggfunc="mean",
    ):
        """Map a property at all mesh vertices to skeletons, aggregating mesh vertices as desired.

        Parameters
        ----------
        mesh_property : array
            N-element array with same length as mesh vertices
        aggfunc : str or function, optional
            Either a pre-defined aggregation function or one of "mean", "median", "max, or "min". By default 'mean'.

        Returns
        -------
        np.ndarray
            Array of aggregated values
        """
        if aggfunc == "mean":
            aggfunc = np.mean
        elif aggfunc == "median":
            aggfunc = np.median
        elif aggfunc == "max":
            aggfunc = np.max
        elif aggfunc == "min":
            aggfunc = np.min
        elif isinstance(aggfunc, str):
            raise ValueError(
                "Only string values allowed are 'mean', 'median', 'max', and 'min'."
            )

        arr = np.vstack((self.skeleton.mesh_to_skel_map, mesh_property)).T
        arr = arr[np.argsort(arr[:, 0])]
        arr_grouped = np.split(
            arr[:, 1], np.unique(arr[:, 0], return_index=True)[1][1:]
        )
        return np.array([aggfunc(x) for x in arr_grouped])

    @property
    @OnlyIfSkeleton.exists
    def branch_points_skel(self):
        """Skeleton index of branch points"""
        return self.SkeletonIndex(self.skeleton.branch_points)

    @property
    @OnlyIfSkeleton.exists
    def branch_points_region(self):
        """List of arrays of all mesh indices associated with each branch point."""
        return self.branch_points_skel.to_mesh_region

    @property
    @OnlyIfSkeleton.exists
    def branch_points(self):
        """Array with one mesh point per skeleton branch point."""
        return self.branch_points_skel.to_mesh_region_point

    @property
    @OnlyIfSkeleton.exists
    def end_points_skel(self):
        """Skeleton index of each end point"""
        return self.SkeletonIndex(self.skeleton.end_points)

    @property
    @OnlyIfSkeleton.exists
    def end_points(self):
        """Array with one mesh point per skeleton end point."""
        return self.end_points_skel.to_mesh_region_point

    @property
    @OnlyIfSkeleton.exists
    def end_points_region(self):
        """List of arrays of all mesh indices associated with each end point."""
        return self.end_points_skel.to_mesh_region

    @property
    @OnlyIfSkeleton.exists
    def root_skel(self):
        """Skeleton index of the root node (as a single element array)"""
        return self.SkeletonIndex([self.skeleton.root])

    @property
    @OnlyIfSkeleton.exists
    def root_region(self):
        """Array of all mesh indices associated with the root node."""
        return self.root_skel.to_mesh_region[0]

    @property
    @OnlyIfSkeleton.exists
    def root(self):
        """Mesh index of one (arbitrary) point associated with the root node."""
        return self.root_skel.to_mesh_region_point[0]

    @OnlyIfSkeleton.exists
    def parent_index(self, mesh_inds, include_parent_free=False, return_as_skel=False):
        """Given a list of mesh indicies, get mesh indices of parents.

        Parameters
        ----------
        mesh_inds : array or int
            Mesh indices to check
        include_parent_free : bool, optional
            If True, returns one element for each mesh index. Nodes with no parents have a -1.
            If False (default), returns a list of elements that omits elements without parents,
            breaking a one-to-one correspondance with the inputs.
        return_as_skel : bool, optional
            Return as skeleton indices. This effectively includes parent-free nodes as well.

        Returns
        -------
        array
            Collection of parent indices in the desired form
        """
        mesh_inds = self._convert_to_meshindex(mesh_inds)
        parent_index = self.skeleton.parent_nodes(mesh_inds.to_skel_index)
        if return_as_skel:
            return parent_index
        if include_parent_free:
            return parent_index.to_mesh_region_point
        else:
            return parent_index[parent_index >= 0].to_mesh_region_point

    @OnlyIfSkeleton.exists
    def child_index(self, mesh_inds, return_as_skel=False):
        """Given a list of mesh indicies, get list of mesh indices of child nodes.

        Parameters
        ----------
        mesh_inds : array or int
            Mesh indices to check
        return_as_skel : bool, optional
            Return as skeleton indices.

        Returns
        -------
        list
            List of arrays of child indices in the desired form
        """
        if np.isscalar(mesh_inds):
            return_scalar = True
        else:
            return_scalar = False
        mesh_inds = self._convert_to_meshindex(mesh_inds)
        child_index = self.skeleton.child_nodes(mesh_inds.to_skel_index)
        if return_as_skel:
            return child_index
        if return_scalar:
            return child_index[0].to_mesh_region_point
        return [n.to_mesh_region_point for n in child_index]

    @OnlyIfSkeleton.exists
    def jump_proximal(self, ind, include_initial=False, hops=1):
        """Find the next branch point (or root) towards root from a given index.

        Parameters
        ----------
        ind : int or MeshIndex
            Mesh index t
        include_initial : bool, optional
            If False, ind is never included in the list of branch points toward root.
            Default is False.
        hops : int, optional
            Number of hops toward root to take. Default is 1.

        Returns
        -------
        mesh_index
            Index of the proximal branch point/root after the desired number of hops.
        """
        ind = self._convert_to_meshindex(ind)
        skind = ind.to_skel_index[0]

        topo_points = self.skeleton.topo_points
        if include_initial == False:
            if skind != self.skeleton.root:
                topo_points = topo_points[topo_points != ind]

        ptr = self.skeleton.path_to_root(skind)
        is_topo = np.isin(ptr, self.skeleton.topo_points)
        topo_pts_on_path = ptr[np.flatnonzero(is_topo)]
        hop_ind = np.min((len(topo_pts_on_path), hops)) - 1
        skind_out = self.SkeletonIndex(topo_pts_on_path[int(hop_ind)])
        return skind_out.to_mesh_region_point

    @OnlyIfSkeleton.exists
    def jump_distal(self, ind):
        """Finds the next topologically interesting (branch or end point) away from root starting from an initial mesh index.

        Parameters
        ----------
        ind : Mesh index
            Index of the initial mesh to use for

        Returns
        -------
        mesh_index
            Index of a mesh point corresponding to the jump.
        """
        ind = self._convert_to_meshindex(ind)
        skind = ind.to_skel_index
        if skind in np.concatenate(
            (self.skeleton.end_points, self.skeleton.branch_points)
        ):
            return ind  # If it's an end point, return immediately

        d = sparse.csgraph.dijkstra(
            self.skeleton.csgraph.T, directed=True, indices=skind
        )

        topo_pts = self.skeleton.topo_points
        proximal_pt = topo_pts[np.argmin(d[0, topo_pts])]
        return self.SkeletonIndex(proximal_pt).to_mesh_region_point

    @OnlyIfSkeleton.exists
    def distance_to_root(self, mesh_indices=None):
        """Distance to root for mesh indices along skeleton.

        Parameters
        ----------
        mesh_indices : array-like
            Collection of mesh indices to look up

        Returns
        -------
        array
            Array of distances to root measured along the skeleton. If no corresponding
            skeleton index exists for a mesh point, a NaN is used.
        """
        if mesh_indices is None:
            mesh_indices = np.arange(len(self.mesh.vertices))
        mesh_indices = self._convert_to_meshindex(mesh_indices)
        ds = np.full(len(mesh_indices), np.nan)
        skinds = mesh_indices.to_skel_index_padded
        ds[skinds >= 0] = self.skeleton.distance_to_root[skinds[skinds >= 0]]
        return ds

    @OnlyIfSkeleton.exists
    def downstream_of(self, mesh_index, return_as_skel=False):
        """Get mesh indices distal to a collection of mesh indices

        Parameters
        ----------
        mesh_index : int or array-like
            Mesh index or collection of mesh indices
        return_as_skel : bool, optional
            If True, returns downstream indices as SkeletonIndex. Default is False.

        Returns
        -------
        array or list
            If a single index was given, returns an array of mesh indices downstream.
            If a collection of indices were given, returns a list of arrays of mesh indices downstream.
        """
        if np.isscalar(mesh_index):
            use_scalar = True
            mesh_index = [mesh_index]
        else:
            use_scalar = False
        mesh_index = self._convert_to_meshindex(mesh_index)
        skinds_downstream = self.skeleton.downstream_nodes(mesh_index.to_skel_index)
        if return_as_skel:
            return skinds_downstream
        minds_downstream = []
        for ds_list in skinds_downstream:
            minds_downstream.append(ds_list.to_mesh_index)
        if use_scalar:
            minds_downstream = minds_downstream[0]
        return minds_downstream

    def _get_segment_list(self, seg_inds, return_as_skel):
        segment_list = []
        for seg in seg_inds:
            if return_as_skel is False:
                segment_list.append(self.skeleton.segments[seg].to_mesh_index)
            else:
                segment_list.append(self.skeleton.segments[seg])
        return segment_list

    @OnlyIfSkeleton.exists
    def same_segment(self, mesh_inds, return_as_skel=False):
        """Get all indices within the same segment (region between branch/end points)

        Parameters
        ----------
        mesh_inds : int or array-like
            Mesh index or collection of mesh indices
        return_as_skel : bool, optional
            If True, return as SkeletonIndex. Default is False.

        Returns
        -------
        array or list
            If a single index was given, returns an array of mesh indices in segment.
            If a collection of indices were given, returns a list of arrays of mesh indices in segment.
        """
        if np.isscalar(mesh_inds):
            return_scalar = True
        else:
            return_scalar = False

        mesh_inds = self._convert_to_meshindex(mesh_inds)
        segs = self.skeleton.segment_map[mesh_inds.to_skel_index]
        segment_list = self._get_segment_list(segs, return_as_skel)

        if return_scalar:
            segment_list = segment_list[0]
        return segment_list

    @OnlyIfSkeleton.exists
    def segments(self, return_as_skel=False):
        """Get all segments (regions between topological points)

        Parameters
        ----------
        return_as_skel : bool, optional
            Description

        Returns
        -------
        TYPE
            Description
        """
        return self._get_segment_list(
            np.arange(len(self.skeleton.segments)), return_as_skel
        )

    @OnlyIfSkeleton.exists
    def same_branch(self, ind, hops=1, extend_to_root=False, return_as_skel=False):
        """Get all indices on the same segment (or, optionally, primary branch)
        and all downstream nodes.

        Parameters
        ----------
        inds : int or SkeletonIndex or MeshIndex
            Mesh index to base branch index off of. Must not be an index of the root.
        hops : int, optional
            Number of branch points to hop towards root before finding downstream vertices, by default 1.
        extend_to_root : bool, optional
            If True, starts the branch at the root intead of a branch point, by default False
        return_as_skel : bool, optional
            If True, returns skeleton indices instead of mesh indices, by default False

        Returns
        -------
        branch_inds : MeshIndex or SkeletonIndex
            Indices on the same branch (SkeletonIndex if return_as_skel==True)
        """
        ind = self._convert_to_meshindex(ind)

        if ind in self.root_region:
            raise ValueError("Index is in root, branch is not defined")

        skind = ind.to_skel_index
        if extend_to_root:
            base_ind = self.root
        else:
            base_ind = self.jump_proximal(ind, hops=hops)

        ptb = self.path_between(ind, base_ind, return_as_skel=True)
        source_ind = self.SkeletonIndex(ptb[-2])
        return self.downstream_of(
            source_ind.to_mesh_region_point[0], return_as_skel=return_as_skel
        )

    def _distance_between(self, inds_source, inds_target, graph, squeeze):
        ds = sparse.csgraph.dijkstra(graph, directed=False, indices=inds_source)
        if squeeze:
            return ds[:, inds_target].squeeze()
        else:
            return ds[:, inds_target]

    @OnlyIfSkeleton.exists
    def distance_between(self, inds_source, inds_target, along_path=True, squeeze=True):
        """Get distance matrix between source and target mesh indices along the object

        Parameters
        ----------
        inds_source : int or array
            mesh indices for one side of the paths
        inds_target : int or array
            mesh indices for the other side of the paths
        along_path : bool, optional
            If True (default), use the skeleton
        squeeze : bool, optional
            If True (default), squeezes singlet dimensions of the distance matrix.
        Returns
        -------
        array
            Array of distances with shape (len(inds_source), len(inds_target)) before optional squeezing.
        """
        inds_source = self._convert_to_meshindex(inds_source)
        inds_target = self._convert_to_meshindex(inds_target)

        if along_path:
            return self._distance_between(
                inds_source.to_skel_index_padded,
                inds_target.to_skel_index_padded,
                self.skeleton.csgraph,
                squeeze=False,
            )
        else:
            return self._distance_between(
                inds_source, inds_target, self.mesh.csgraph, squeeze=squeeze
            )

    @OnlyIfSkeleton.exists
    def path_between(self, source_index, target_index, return_as_skel=False):
        """Get indices of a path between a source and a target mesh index.

        Parameters
        ----------
        source_index : int
            Mesh index of a source
        target_index : int
            Mesh index of a target
        return_as_skel : bool, optional
            If True, return as skeleton indices. Default is False.

        Returns
        -------
        array
            All indices between source and target.
        """
        source_index = self._convert_to_meshindex(source_index)
        target_index = self._convert_to_meshindex(target_index)
        skpath = self.SkeletonIndex(
            self.skeleton.path_between(
                source_index.to_skel_index[0], target_index.to_skel_index[0]
            )
        )
        if return_as_skel:
            return skpath
        return skpath.to_mesh_index

    def _within_distance(self, inds, graph, max_distance):
        ds = sparse.csgraph.dijkstra(graph, indices=inds, directed=False)
        return ds < max_distance

    @OnlyIfSkeleton.exists
    def within_distance(
        self, source_inds, distance, collapse=True, return_as_skel=False
    ):
        """Mesh indices within a distance along the object's skeleton

        Parameters
        ----------
        source_inds : array or int
            Source indices of the mesh
        distance : numeric
            Distance to query
        collapse : bool, optional
            If True, returns a mask that is True if index is close to any source. Default is True.
        return_as_skel : bool, optional
            If True, returns skeleton indices. Default is False.

        Returns
        -------
        array or list of arrays
            Mask (if return_as_skel is False) or skeleton index array.
            If collapse is True, a single array that is True if True for any source.
            If collapse is False, a list of arrays for each source.
        """
        if np.isscalar(source_inds) or collapse:
            return_scalar = True
        else:
            return_scalar = False

        source_inds = self._convert_to_meshindex(source_inds)

        dmask = self._within_distance(
            source_inds.to_skel_index, self.skeleton.csgraph, distance
        )
        if collapse:
            if (dmask.shape) == 1:
                dmask = dmask.reshape(1, -1)
            dmask = np.any(dmask, axis=0).reshape(1, -1)
            return_scalar = True
        if return_as_skel:
            if return_scalar:
                dmask = self.SkeletonIndex(np.flatnonzero(dmask))
            else:
                dmask = [self.SkeletonIndex(np.flatnonzero(m)) for m in dmask]
        else:
            dmask = dmask[:, self.skeleton.mesh_to_skel_map]
            if return_scalar:
                dmask = self.MeshIndex(np.flatnonzero(dmask))
            else:
                dmask = [self.MeshIndex(np.flatnonzero(m)) for m in dmask]
        return dmask

    @OnlyIfSkeleton.exists
    def path_length(self, inds=None):
        """Get path length of collection of mesh indices

        Parameters
        ----------
        inds : array-like
            Mesh indices to compute path length for. Can be in any order.

        Returns
        -------
        path_length : float
            Path length in mesh units.
        """
        if inds is None:
            return self.skeleton.path_length()
        inds = self._convert_to_meshindex(inds)
        return self.skeleton.path_length(inds.to_skel_index)

    @OnlyIfSkeleton.exists
    def total_path_length(self):
        return self.skeleton.csgraph.sum()

    @OnlyIfSkeleton.exists
    def linear_density(
        self,
        inds,
        width,
        weight=None,
        normalize=True,
        exclude_root=False,
    ):
        """Compute a sliding window average linear density of points across the object

        Parameters
        ----------
        inds : array
            Mesh indices for density (e.g. synapse locations).
        width : numeric
            width of average window (in all directions).
        weight : array None, optional
            Weight for each point for weighted average. If None, assumes weight of unity.
        normalize : bool, optional
            If False, sums the weights but does not normalize by amount of cable.
            Default is True.
        exclude_root : bool, optional
            If True, disconnects root from the graph for the case that the root is not
            well-approximated by a line (e.g. a cell body.). The density for those vertices
            will be infinite or nan.
            Default is False.


        Returns
        -------
        density_estimate : array
            N-length array of density at all mesh vertices.
        """
        W = window_matrix(self.skeleton, width)

        inds = self._convert_to_meshindex(inds)
        has_inds = np.full(self.skeleton.n_vertices, 0)
        if weight is None:
            skinds, count = np.unique(inds.to_skel_index, return_counts=True)
            has_inds[skinds] = count
        else:
            for w, skind in zip(weight, inds.to_skel_index):
                has_inds[skind] += w
        item_count = W.dot(has_inds.reshape(-1, 1)).ravel()
        if normalize:
            if exclude_root:
                g = self.skeleton.cut_graph(
                    self.skeleton.child_nodes([self.skeleton.root])[0], directed=False
                )
                len_per = np.array(g.sum(axis=1) / 2).ravel()
            else:
                len_per = np.array(
                    self.skeleton.csgraph_undirected.sum(axis=1) / 2
                ).ravel()

            norm = W.dot(len_per.reshape(-1, 1)).ravel()
            with np.errstate(divide="ignore"):
                rho = item_count / norm
        else:
            rho = item_count
        return rho[self.skeleton.mesh_to_skel_map][self.mesh.node_mask]

    ###k########################
    # Visualization functions #
    ###########################

    def requires_vtk(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if _vtk_loaded:
                return func(*args, **kwargs)
            else:
                return None

        return wrapper

    @requires_vtk
    def mesh_actor(self, **kwargs):
        if self.mesh is not None:
            return trimesh_vtk.mesh_actor(self.mesh, **kwargs)
        else:
            return None

    @requires_vtk
    def anno_point_actor(self, anno_name, query=None, filter_query=None, **kwargs):
        row_filter = np.full(len(self.anno[anno_name].df), True)
        if query is not None:
            qry = self.anno[anno_name].query(query)
            row_filter = np.logical_and(row_filter, qry.row_filter)
        if filter_query is not None:
            qry = self.anno[anno_name].filter_query(filter_query)
            row_filter = np.logical_and(row_filter, qry.row_filter)
        if anno_name in self.anno.table_names:
            return trimesh_vtk.point_cloud_actor(
                self.anno[anno_name].points[row_filter], **kwargs
            )

    @OnlyIfSkeleton.exists
    @requires_vtk
    def skeleton_actor(self, **kwargs):
        if self.skeleton is not None:
            return trimesh_vtk.skeleton_actor(self.skeleton, **kwargs)

    @property
    @OnlyIfSkeleton.exists
    def radius(self):
        return self.skeleton_property_to_mesh(self.skeleton.radius)

    ##########
    # Saving #
    ##########
    def save_meshwork(
        self, filename, overwrite=False, version=meshwork_io.LATEST_VERSION
    ):
        """Save meshwork to hdf5 file.

        Parameters
        ----------
        filename : str
            Filename to save to
        overwrite : bool, optional
            If True, overwrites an existing file. Default is False.
        """
        meshwork_io._save_meshwork(filename, self, overwrite=overwrite, version=version)

    @OnlyIfSkeleton.exists
    def export_to_swc(
        self,
        filename,
        resample_spacing=None,
        apical_label=None,
        axon_label=None,
        dendrite_label=None,
        soma_label=None,
        dendrite_default=True,
        interp_kind="linear",
        tip_length_ratio=0.5,
        radius=None,
        radius_agg="mean",
        header=None,
        scaling=1000,
    ):
        """Export a neuron to SWC, optionally resampling and using annotations to label compartments.

        Parameters
        ----------
        filename : str
            Location where the swc file will be saved
        resample_spacing : float or None, optional
            Desired spacing between skeleton vertices. If None, no resampling is performed. Default is None.
        apical_label : str or array, optional
            If a string, name of annotation spanning the apical dendrite compartment. If a numpy array, one that
            includes all mesh vertices that should be labeled as apical. SkeletonIndex or MeshIndex arrays that
            cover all such apical vertices are also interpreted correctly. By default, None
        axon_label : str or array, optional
            Same as apical_label, but for the axon compartment. by default None.
        dendrite_label : str or array, optional
            Same as apical_label, but for the (basal) dendrite compartment. by default None.
        soma_label : str or array, optional
            Same as apical_label, but for the soma compartment. by default None.
        dendrite_default : bool, optional
            If True, assumes any vertex without a specific label is basal dendrite, by default True.
        interp_kind : str, optional
            Interpolation method used by scipy.interpolate.interp1d, by default "linear".
        tip_length_ratio : float, optional
            Ratio of spacing of branch tip length to desired spacing to keep a final tip, by default 0.5.
        radius : float or array, optional
            Array of radius values for the skeleton (in the same units as vertex coordinates), by default None.
            Array can be either one value per mesh vertex (in which case values are aggregated) or one value per skeleton vertex.
            Alternatively, a single number will be assigned for all vertices.
        radius_agg : str or func, optional
            If mapping mesh radius values to skeleton vertices, how to aggregate values. Options are 'mean', 'median',
            'min', 'max', and a user-defined aggregation function. By default, "mean".
        header : str, optional
            Header string for the SWC file, by default None.
        scaling : float, optional
            Scaling between coordinates of the file and exported coordiates, by default 1000 because for EM
            we usually have coordinates in nanometers and SWC usually expects microns.
        """

        def _extract_annotation(anno_value):
            if anno_value is not None:
                if isinstance(anno_value, str):
                    return self.anno[anno_value].skel_mask
                else:
                    anno_minds = self._convert_to_meshindex(anno_value)
                    return anno_minds.to_skel_mask
            else:
                return None

        apical_inds = _extract_annotation(apical_label)
        axon_inds = _extract_annotation(axon_label)
        dendrite_inds = _extract_annotation(dendrite_label)
        soma_inds = _extract_annotation(soma_label)

        node_labels = swc_node_labels(
            self.skeleton,
            dendrite_indices=dendrite_inds,
            apical_indices=apical_inds,
            soma_indices=soma_inds,
            axon_indices=axon_inds,
            dendrite_default=dendrite_default,
        )

        if radius is not None:
            if np.isscalar(radius):
                radius = np.full(self.skeleton.n_vertices, radius)
            elif len(radius) == self.mesh.n_vertices:
                radius = self.mesh_property_to_skeleton(radius, aggfunc=radius_agg)
            elif len(radius) != self.skeleton.n_vertices:
                raise ValueError(
                    "Radius array must correspond to either mesh or skeleton vertices"
                )

        self.skeleton.export_to_swc(
            filename=filename,
            node_labels=node_labels,
            radius=radius,
            header=header,
            xyz_scaling=scaling,
            resample_spacing=resample_spacing,
            interp_kind=interp_kind,
            tip_length_ratio=tip_length_ratio,
        )


def load_meshwork(filename):
    """Loads meshwork file from an hdf5 file.

    Parameters
    ----------
    filename : str
        file location

    Returns
    -------
    meshwork : Meshwork
        Meshwork object from file
    """
    meta, mesh, skel, annos, mask = meshwork_io._load_meshwork(filename)
    mw = Meshwork(
        mesh,
        skeleton=skel,
        seg_id=meta.get("seg_id", None),
        voxel_resolution=meta.get("voxel_resolution", DEFAULT_VOXEL_RESOLUTION),
    )
    for name, data in annos.items():
        mw.add_annotations(
            name=name,
            data=data.get("data"),
            anchored=data.get("anchor_to_mesh"),
            point_column=data.get("point_column"),
            max_distance=data.get("max_distance"),
            index_column=data.get("index_column", None),
        )
    if not np.all(mask == mesh.node_mask):
        mw.apply_mask(mask)
    return mw
